""" Read GFF files and output isomiRs compatible format"""
from __future__ import print_function

import os

import mirtop.libs.logger as mylog
from mirtop.mirna import fasta, mapper
from mirtop.gff.classgff import feature
from mirtop.gff.header import read_samples
from mirtop.mirna.realign import get_mature_sequence, align_from_variants
from mirtop.mirna.realign import read_id, variant_to_5p, \
                                 variant_to_3p, variant_to_add
from mirtop.gff.body import variant_with_nt

logger = mylog.getLogger(__name__)


def convert(args):
    """
    Main function to convert from GFF3 to isomiRs Bioc Package.

    Reads a GFF file to produces output file containing Expression counts

    Args:
        *args(namedtuple)*: arguments parsed from command line with
            *mirtop.libs.parse.add_subparser_counts()*.

    Returns:
        *file (file)*: with columns like:
            UID miRNA Variant Sample1 Sample2 ... Sample N
    """
    logger.info("INFO Writing TSV file to directory %s", args.out)
    for gff in args.files:
        logger.info("INFO Reading GFF file %s", gff)
        _convert_file(gff, args)


def _convert_file(gff, args):
    sep = "\t"
    precursors = fasta.read_precursor(args.hairpin, args.sps)
    matures = mapper.read_gtf_to_precursor(args.gtf, args.database)
    variant_header = sep.join(['mism', 'add', 't5', 't3'])

    gff_file = open(gff, 'r')
    out_file = os.path.join(args.out, "%s_rawData.tsv" % os.path.splitext(os.path.basename(gff))[0])
    missing_parent = 0
    missing_mirna = 0
    unvalid_uid = 0
    with open(out_file, 'w') as outh:

        for samples_line in gff_file:
            if samples_line.startswith("## COLDATA:"):
                samples = sep.join(samples_line.strip().split("COLDATA:")[1].strip().split(","))
                header = sep.join(['seq', 'mir',
                                   variant_header, samples])
                print(header, file=outh)
                break

        for mirna_line in gff_file:
            gff = feature(mirna_line)
            attr = gff.attributes
            UID = attr["UID"]
            Read = attr["Read"]
            mirna = attr["Name"]
            parent = attr["Parent"]
            variant = attr["Variant"]
            try:
                Read = read_id(UID)
            except KeyError:
                unvalid_uid += 1
                continue

            expression = sep.join(attr["Expression"].strip().split(","))
            cols_variants = sep.join(_expand(variant))
            logger.debug("COUNTS::Read:%s" % Read)
            logger.debug("COUNTS::EXTRA:%s" % variant)
            if parent not in precursors:
                missing_parent += 1
                continue
            if mirna not in matures[parent]:
                missing_mirna += 1
                continue
            extra = variant_with_nt(mirna_line, precursors, matures)
            if extra == "Invalid":
                continue
            logger.debug("COUNTS::EXTRA:%s" % extra)
            cols_variants = sep.join(_expand(extra, True))
            summary = sep.join([Read,  mirna,
                                cols_variants, expression])
            logger.debug(summary)
            print(summary, file=outh)

    gff_file.close()
    logger.info("Missing Parents in hairpin file: %s" % missing_parent)
    logger.info("Missing MiRNAs in GFF file: %s" % missing_mirna)
    logger.info("Non valid UID: %s" % unvalid_uid)
    logger.info("Output file is at %s" % out_file)


def _expand(variant, nts=False):
    """Expand Variant field into list for iso_5p, iso_3p, iso_add3p, iso_snv"""
    list_variant = []
    isomir = {}
    snp_var = []
    for v in variant.split(","):
        if v.find(":") > 0:
            isomir[v.split(":")[0]] = v.split(":")[1]
        elif v.find("snv") > 0:
            snp_var.append(1)
    if nts:
        if "iso_snv" in isomir:
            list_variant.append(isomir["iso_snv"])
        else:
            list_variant.append(0)
    else:
        snp = sum(snp_var)
        list_variant.append(snp)
    if "iso_add3p" in isomir:
        list_variant.append(isomir["iso_add3p"])
    else:
        list_variant.append(0)
    if "iso_5p" in isomir:
        list_variant.append(isomir["iso_5p"])
    else:
        list_variant.append(0)
    if "iso_3p" in isomir:
        list_variant.append(isomir["iso_3p"])
    else:
        list_variant.append(0)

    return list(map(str, list_variant))



from mirtop.gff import gff_versions
from collections import OrderedDict


class feature:
    """"Class with alignment information."""

    def __init__(self, line, sep="="):
        # if isinstance(line, basestring) # str in python 3
        if isinstance(line, dict):
            line = self.create_line(line, sep)
        self.line = line
        self.attributes = {}
        self.columns = {}
        self.read_gff_line()
        self.reserved_attributes = gff_versions.ATTRv['1.1']

    def guess_format(self):
        return "=" if self.line.find("Name=") > -1 else " "

    def create_line(self, fields, sep):
        """
        Create GFF/GTF line from dictionary with fields
        """
        # TODO: need unit test. Check all exists.
        fields["attrb"] = ("Read {seq_name}; UID {idseq}; Name {name};"
                           "Parent {parent}; "
                           "Variant {variant}; Cigar {cigar};"
                           "Expression {counts}; "
                           "Filter {filter}; Hits {hits};").format(**fields)

        line = ("{chrom}\t{database}\t{source}\t{start}\t{end}"
                "\t{score}\t{strand}\t.\t{attrb}").format(**fields)
        return line

    def paste_columns(self, sep=None):
        """
        Create GFF/GTF line from read_gff_line
        """
        sep = self.guess_format()
        attributes = "%s;" % "; ".join(
            "%s%s%s" % (a, sep, self.attributes[a]) for a in self.attributes)
        return "\t".join([self.columns['chrom'], self.columns['source'],
                          self.columns['type'],
                          self.columns['start'], self.columns['end'],
                          self.columns['score'],
                          self.columns['strand'], self.columns['ext'],
                          attributes])

    def read_attributes(self, gff_attrb, sep=" "):
        """
        Read GFF/GTF attributes and return dictionary with fields
        """
        sep = self.guess_format()
        gff_dict = OrderedDict()
        for gff_item in gff_attrb.split(";"):
            item_pair = gff_item.strip().split(sep)
            if len(item_pair) > 1:
                gff_dict[item_pair[0].strip()] = item_pair[1].strip()
        self.attributes = gff_dict

    def read_gff_line(self):
        """
        Read GFF/GTF line and return dictionary with fields
        """
        line = self.line
        if line.startswith("#"):
            return line
        cols = line.strip().split("\t")
        if len(cols) != 9:
            raise ValueError("Line has less than 9 elements: %s" % line)
        self.read_attributes(cols[8])
        fields = {'chrom': cols[0],
                  'source': cols[1],
                  'type': cols[2],
                  'start': cols[3],
                  'end': cols[4],
                  'score': cols[5],
                  'strand': cols[6],
                  'ext': cols[7]}
        self.columns = fields



# -*- coding: utf-8 -*-

"""
Function that loads a gff file into a pandas dataframe
"""
import pandas as pd
import numpy as np


def loadfile(filename,verbose=True):
        
    try:
        if verbose==True:
            print('Loading', filename)
        # obtaning sample names and number from 3rd line in header
        num_header_lines=0
        with open(filename) as f:
            rowfile=f.readline()
            num_header_lines+=1
            while True:
                if rowfile.startswith('## COLDATA'):
                    sample_names=rowfile.split()[2].split(',')
                    break
                else:
                    rowfile=f.readline()
                    num_header_lines+=1
        sample_number = len(sample_names)
        if verbose==True:
            print('--------------------------------------')
            print(sample_number,' samples in the file')
            print('--------------------------------------')
            for elem in sample_names:
                print(elem)
            print('--------------------------------------')
        
        
        
        
        # number of columns in gff file
        gff_cols = pd.read_table(filename, sep='\t', skiprows=num_header_lines, header=None).columns
         
        # Adquiring non-attributes data
        body_data=pd.read_table(filename, sep='\t', skiprows=num_header_lines, header=None, usecols=gff_cols[0:-1])
        body_data.columns = ['SeqID', 'source', 'type', 'start', 'end', 'score', 'strand', 'phase']
        # Adquiring attributes data
        atr_data = pd.read_table(filename, sep='\t', skiprows=num_header_lines, header=None, usecols=gff_cols[[-1]])
        
        #print 'hasta aqui todo bien'
        # Splitting the attributes column
        list_atr = []
        # cheking attributes present in first row
 
        
        attr_names=[attr.split()[0] for attr in atr_data.values[0,0].split(';')]
        #attributes in the column
        
      
        #print attr_names
        num_attr = len(attr_names)                         #number of attributes
        #expression_colindex=attr_names.index ('Expression')  #position of the expression column in the attr column 
        if verbose==True:
            print(num_attr,' attributes in the file ')
            print('--------------------------------------')
            for attr in attr_names:
                print(attr)
            print('--------------------------------------')
        
        # joining rows of attributes without the descriptor
        for row in range(atr_data.shape[0]):
            list_atr.append([attr.split()[1] for attr in atr_data.values[row,0].split(';')])
            #list_atr.append(atr_data.values[row, 0].split()[1::2])
    
        # appending observations
        atr_data = pd.DataFrame(list_atr, columns=attr_names)
   
        # desglosing the expression column in a column for each sample
        list_expression=[]
        for row in range(atr_data.shape[0]):
            list_expression.append(atr_data.loc[row,'Expression'].split(','))
        sample_names=['Expression_' +x for x in sample_names]
        
        expression_data = pd.DataFrame(list_expression, columns=sample_names)      
        
        
        atr_data=atr_data.drop('Expression',axis=1) #Remove the expression column
        atr_data=atr_data.join(expression_data)
    
        # Joining the body and attributes dataframes
        data = body_data.join(atr_data)
        
        #Unlisting the variant column
        tempframe=data[data.type=='isomiR']
        list_variants_present=[]
        for row in tempframe.itertuples():      
            actual_variant=tempframe.loc[row.Index,'Variant'].split(',')
            for i in range(len(actual_variant)):
                if actual_variant[i] not in list_variants_present:
                    list_variants_present.append(actual_variant[i])
                    
                    
        
        for var in list_variants_present:
            for row in data.itertuples():
                
                try:
                    index=data.loc[row.Index,'Variant'].split(',').index(var)
                except:
                    index=-1
                if index>=0:
                    #print var, data.loc[row.Index,'Variant']
                    data.at[row.Index,var]=1
                else:
                    data.at[row.Index,var]=np.nan    
        return data
    except:
        print('Error loading the file')


"""
Function that check the header then load a gff file and check the content
Returns the dataframe if the format is ok, false if not
"""
def load_check_gff3(filename):
    try:
        Error = False
        coldata_found=False
        # Checking the format file 
        # Header and 1st data row
        with open(filename) as file:
            rowfile=file.readline()           
            while True:
                if rowfile.startswith('##'):
                    if rowfile.startswith("## COLDATA"):
                         coldata_found=True
                    rowfile=file.readline()
                else:
                    data_1=rowfile.split('\t')
                    break
        if coldata_found==False:
            print('No COLDATA, bad header')
            return False
        
        #Number of columns without breaking down attributes column 
        if len(data_1) > 9:
            Error=True
            print(len(data_1))
            print('Too much columns')
        # Cheking the attributes column  
        attr_names=data_1[-1].split(';')
        list_attr=[]
        for atr in range(len(attr_names)-1):
            list_attr.append(attr_names[atr].split()[0])
        possible_attr = ['UID', 'Read', 'Name', 'Parent', 'Variant', 'Cigar', 'Hits', 'Alias', 'Genomic', 'Expression',
                           'Filter', 'Seed_fam']
        for attr in list_attr:    
            if attr not in possible_attr:
                Error=True
                print(attr,'is not a possible attribute')
                break  
        if Error:
            print('File format error')
            return False      
        # If not format error, loading content
        try:
            dataframe=loadfile(filename,True)
        except:
            print('Error loading file')
            return False
        print('Checking content')
        for i in range(dataframe.shape[0]):
            # Labels in type column
            if dataframe.loc[i, 'type'] not in ['ref_miRNA', 'isomiR']:
                Error = True
                print('line', i, 'pip install Markdownbad type error')
    
            # start<end
            if dataframe.loc[i, 'start'] >= dataframe.loc[i, 'end']:
                Error = True
                print('line', i, 'start >=end error')
    
            # Strand + or -
            if dataframe.loc[i, 'strand'] not in ['+', '-']:
                Error = True
                print('line', i, 'bad strand error')
            # Variant checking
            possible_variant=['iso_5p','iso_3p','iso_add','iso_snp_seed','iso_snp_central_offset','iso_snp_central',
                  'iso_central_supp','iso_snp_central_supp','iso_snp']
            
            variant_i=dataframe.loc[i,'Variant'].split(',')
            
            if len(variant_i)==1 and variant_i[0]!='NA':
                if variant_i[0].split(':')[0] not in possible_variant:
                    Error = True
                    print('Variant error', variant_i[0].split(':')[0], 'line', i)
            elif variant_i[0]!='NA':
                for var in range(len(variant_i)):
                    if variant_i[var].split(':')[0] not in possible_variant:
                        Error = True
                        print('Variant error', variant_i[0].split(':')[0], 'line', i)

        #Checking expression data
        expression_cols=[col for col in dataframe.columns if 'Expression_' in col]         
        for col in expression_cols:
            for i in range(dataframe.shape[0]):
                if not dataframe.loc[i,col].isdigit():
                    print(dataframe.loc[i,col].isdigit())
                    print('Expression count error line',i)
                    Error= True
            dataframe[col]=dataframe[col].astype(int) #setting the datatype of counts
            dataframe[col]=dataframe[col].replace(0,np.nan) #Setting 0 reads to NaN
        if 'Filter' in dataframe.columns:         
            for i in range(dataframe.shape[0]):
                if dataframe.loc[i, 'Filter']!='Pass':
                    print('Warning non-pass filter in line',i)
        if Error:
            print('File format error')
            return False

        print('--------------------------------------')
        print(dataframe.dtypes)
        print('--------------------------------------')
        print('Format ok')
        return dataframe
    except:
        print('Error checking the file')
        return False



from __future__ import print_function

import argparse
import os
import random
from collections import defaultdict

import pysam

import mirtop.libs.logger as mylog
import mirtop.libs.do as runner


def _read_fasta(fa, size):
    source = dict()
    with open(fa) as inh:
        for line in inh:
            if line.startswith(">"):
                name = line.strip().split()[0].replace(">", "")
            else:
                if len(line.strip()) >= size:
                    source.update({name: line.strip()[0:size]})
    return source


def _update_ends(source):
    nts = ["A", "T", "C", "G"]
    start_idx = 0
    end_idx = 0
    for name in source:
        source[name] = nts[start_idx] + source[name] + nts[end_idx]
        if end_idx == 3 and start_idx == 3:
            end_idx = -1
            start_idx = 0
        if end_idx == 3:
            start_idx += 1
            end_idx = 0
        end_idx += 1
    return source


def _write_fasta(sequences, filename):
    with open(filename, 'w') as outh:
        for name in sequences:
            if sequences[name]:
                print(">%s\n%s" % (name, sequences[name]), file=outh)
    return filename


def _parse_hits(sam, source):
    uniques = defaultdict(list)
    # read sequences and score hits (ignore same sequence)
    handle = pysam.Samfile(sam, "rb")
    for line in handle:
        reference = handle.getrname(line.reference_id)
        name = line.query_name
        # sequence = line.query_sequence if not line.is_reverse else reverse_complement(line.query_sequence)
        if reference == name:
            continue
        # print([reference, name, line.get_tag("NM")])
        distance = line.get_tag("NM")
        uniques[name].append(distance)
        uniques[reference].append(distance)
    # read parsed data and keep the ones with score > 10 edit distance
    for name in uniques:
        if min(uniques[name]) < 4:
            if name in source:
                source[name] = None
    return source


parser = argparse.ArgumentParser()
parser.add_argument("--fa",
                    help="File with mature sequences.", required=True)
parser.add_argument("-s", "--size", default=22,
                    help="Size of spike-ins to generate.")
parser.add_argument("-n", "--number", default=16,
                    help="Number of spike-ins to generate.")
parser.add_argument("-o", "--out", default="spikeins.fa",
                    help="Name used for output files.")
parser.add_argument("--seed", help="set up seed for reproducibility.",
                    default=42)
parser.add_argument("--universe", help="Set up universe sequences to avoid duplication.",
                    default=None)

args = parser.parse_args()
random.seed(args.seed)

mylog.initialize_logger(os.path.dirname(os.path.abspath(args.out)))
logger = mylog.getLogger(__name__)

# Read file to get all sequences longer than size - 2
size = args.size - 2
source = _read_fasta(args.fa, size)
logger.info("%s was read: %s sequences were loaded" % (args.fa, len(source)))
source = _update_ends(source)
logger.info("source updated with extended nts: %s" % source)

# Map all vs all with razers3
modified = _write_fasta(source, os.path.join(os.path.dirname(args.out), "modified.fa"))
sam = os.path.join(os.path.dirname(args.out), "modified.bam")
runner.run(("razers3 -i 75 -rr 80 -f -so 1 -o {output} {target} {query}").format(output=sam, target=modified, query=modified))
uniques = _parse_hits(sam, source)
print(uniques)
if args.universe:
    sam = os.path.join(os.path.dirname(args.out), "modified_vs_universe.sam")
    runner.run(("razers3 -i 75 -rr 80 -f -o {output} {target} {query}").format(output=sam, target=args.universe, query=modified))
    uniques = _parse_hits(sam, uniques)
print(uniques)

# Write uniques to fasta
_write_fasta(uniques, args.out)



from __future__ import print_function

import datetime
import sys
import os.path as op

import six

from mirtop.mirna.fasta import read_precursor
from mirtop.mirna.mapper import read_gtf_to_precursor, read_gtf_to_mirna
from mirtop.gff.body import read_gff_line
import mirtop.libs.logger as mylog

logger = mylog.getLogger(__name__)


def convert(args):
    """
    Main function to convert from GFF3 to VCF.

    Args:
      *args*: supported options for this sub-command.
        See *mirtop.libs.parse.add_subparser_export()*.
    """
    for fn in args.files:
        out_file = op.join(args.out, "%s.vcf" % op.splitext(op.basename(fn))[0])
        logger.info("Reading %s" % fn)
        create_vcf(fn, args.hairpin, args.gtf, out_file, args.database)
        logger.info("VCF generated %s" % out_file)


def cigar_2_key(cigar, readseq, refseq, pos, var5p, var3p, parent_ini_pos, parent_end_pos, hairpin):
    """
    Args:
        'cigar(str)': CIGAR standard of a compressed alignment representation, this CIGAR omits the '1' integer.
        'readseq(str)': the read sequence
        'refseq(str)': the reference sequence
        'pos(str)': the start current position
        'var5p(int)': extra nucleotides not in the reference miRNA (5p strand)
        'var3p(int)': extra nucleotides not in the reference miRNA (3p strand)
        'parent_ini_pos(int)': the start position of the parent miRNA
        'parent_end_pos(int)': the last position of the parent miRNA
        'hairpin(str)': the string of the hairpin for all the miRNA
    Returns:
        'key_pos(str list)': a list with the positions of the variants.
        'key_var(str list)': a list with the variant keys found.
        'ref(str)': reference base(s).
        'alt(str)': altered base(s).
    """
    key_pos = []
    key_var = []
    ref = []
    alt = []
    n_Mpar = "0"
    n_M = 0
    n_NM = 0
    n_D = 0
    n_I = 0
    # Obtaining the part of the hairpin outside of its parent region (previous)
    if var5p < 0:
        previous_pos = hairpin[parent_ini_pos + var5p:parent_ini_pos]
        refseq = previous_pos + refseq
    else:
        refseq = refseq[var5p:]
    # Obtaining the part of the hairpin outside of its parent region (posterior)
    if var3p >= 0:
        posterior_pos = hairpin[parent_end_pos+1:parent_end_pos+var3p+1]
        refseq = refseq + posterior_pos
    else:
        refseq = refseq[:var3p]
    # Calculating the type of variant and its ref/alt positions.
    logger.debug("VCF::precursor::read %s" % refseq)
    logger.debug("VCF::CIGAR::read %s" % cigar)
    for i in range(len(cigar)):
        if cigar[i].isdigit():
            n_Mpar = n_Mpar + (cigar[i])
        elif cigar[i] in ["A", "T", "C", "G"]:
            pos_rel_ref = n_M + n_D + n_NM + 1
            key_pos.append(pos + pos_rel_ref - 1)
            key_var.append(cigar[i])
            ref.append(refseq[pos_rel_ref - 1])
            alt.append(cigar[i])
            n_NM += 1
        elif cigar[i] == "D":
            if i == 0:
                print("Unexpected error: 'D' in first Cigar pos")
            elif cigar[i-1] == "D":
                ref[-1] = ref[-1] + refseq[n_M + n_NM + n_D]  # Adds new del in the REF column
                key_var[-1] = "D" + str(int(key_var[-1][1:]) + 1)  # Adds '1' to the number of Dels in succession
                n_D += 1
            else:
                pos_rel_ref = n_M + n_NM + n_D
                key_pos.append(pos + pos_rel_ref)
                ref.append(refseq[pos_rel_ref-1:pos_rel_ref+1])
                alt.append(refseq[pos_rel_ref-1])
                key_var.append("D1")
                n_D += 1
        elif cigar[i] == "I":
            if i == 0:
                print("Unexpected error: 'I' in first Cigar pos")
            elif cigar[i-1] == "I":
                alt[-1] = alt[-1] + readseq[n_M + n_NM + n_I]  # Adds the new Insert in the ALT column
                key_var[-1] = key_var[-1] + 'I' + readseq[n_M + n_NM + n_I]  # Adds new Ins in the Key
                n_I += 1
            else:
                pos_rel_read = n_M + n_NM + n_I
                pos_rel_ref = n_M + n_NM + n_D
                key_pos.append(pos + pos_rel_ref)
                alt.append(readseq[pos_rel_read-1:pos_rel_read+1])
                ref.append(refseq[pos_rel_ref-1])
                key_var.append("I" + alt[-1][-1])
                n_I += 1
        elif cigar[i] == "M":
            if (i == 0) | (not cigar[i - 1].isdigit()):
                n_M = n_M + 1
            else:
                n_M = n_M + int(n_Mpar)
                n_Mpar = "0"
        else:
            print("Unexpected error: %s, index: %s" % (cigar[i], i))
    return(key_pos, key_var, ref, alt)


def create_vcf(mirgff3, precursor, gtf, vcffile, database):
    """
    Args:
        'mirgff3(str)': File with mirGFF3 format that will be converted
        'precursor(str)': Fasta format sequences of all miRNA hairpins
        'gtf(str)': Genome coordinates
        'vcffile': name of the file to be saved
    Returns:
        Nothing is returned, instead, a VCF file is generated
    """
    #Check if the input files exist:
    try:
        gff3_file = open(mirgff3, "r", encoding="utf-8") if six.PY3 else open(mirgff3, "r")
    except IOError:
        print ("Can't read the file", end=mirgff3)
        sys.exit()
    with gff3_file:
        data = gff3_file.read()
        if six.PY2:
            data = data.decode("utf-8-sig").encode("utf-8")

    gff3_data = data.split("\n")
    vcf_file = open(vcffile, "w")

    ver = "v4.3"  # Current VCF version formatting
    vcf_file.write("##fileformat=VCF%s\n" % ver)
    date = datetime.datetime.now().strftime("%Y%m%d")
    vcf_file.write("##fileDate=%s\n" % date)
    source = "\n".join(s for s in gff3_data if "## source-ontology: " in s)[20:]
    line = 0
    sample_names = []
    while gff3_data[line][:2] == "##":
        if gff3_data[line][:19] == "## source-ontology:":
            source = gff3_data[line][20:]
        elif gff3_data[line][:11] == "## COLDATA:":
            sample_names = gff3_data[line][12:].split(",")
        line += 1
    vcf_file.write("##source=%s\n" % source)
    vcf_file.write('##INFO=<ID=NS,Type=Integer,Description="Number of samples"\n')
    vcf_file.write("##FILTER=<ID=REJECT,Description='"'Filter not passed'"'>\n")
    vcf_file.write('##FORMAT=<ID=TRC,Number=1,Type=Integer,Description="Total read count">\n')
    vcf_file.write('##FORMAT=<ID=TSC,Number=1,Type=Integer,Description="Total SNP count">\n')
    vcf_file.write('##FORMAT=<ID=TMC,Number=1,Type=Integer,Description="Total miRNA count">\n')
    vcf_file.write('##FORMAT=<ID=GT,Number=1,Type=Integer,Description="Genotype">\n')
    header = "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT"
    # Adds Header
    for s in range(len(sample_names)):
        header = header + "\t" + sample_names[s]
    vcf_file.write(header)

    all_dict = dict()  # initializing an empty dictionary where all info will be added
    key_list = []  # Initializing a list which will contain all the keys of the dictionary
    mirna_dict = dict()  # initializing an empty dictionary where mirna info will be put
    n_SNP = 0
    n_noSNP = 0
    no_var = 0
    hairpins = read_precursor(precursor)
    gff3 = read_gtf_to_precursor(gtf, database)
    gtf_dic = read_gtf_to_mirna(gtf)
    for line in range(0, len(gff3_data)):
        if not gff3_data[line]:
            continue
        if gff3_data[line][1] == "#":
            continue
        else:   # Parsing the gff3 mirna lecture:
            gff_fields = read_gff_line(gff3_data[line])
            gtf_name = gff_fields['attrb']['Name']
            gtf_parent = gff_fields['attrb']['Parent']
            if gtf_parent not in gff3:
                continue
            if gtf_name not in gff3[gtf_parent]:
                    continue
            parent_ini_pos = gff3[gtf_parent][gtf_name][0]
            parent_end_pos = gff3[gtf_parent][gtf_name][1]
            ref_seq = (hairpins[gtf_parent][parent_ini_pos:parent_end_pos+1])
            vcf_chrom = gtf_dic[gtf_name][gtf_parent][0]
            vcf_pos = int(gff_fields['start']) + int(gtf_dic[gtf_name][gtf_parent][1])
            hairpin = hairpins[gtf_parent]
            variants = gff_fields['attrb']['Variant'].split(",")
            logger.debug("VCF::Variant::%s" % variants)
            #  Obtaining the iso_3p, iso_add3p and iso_5p values:

            var3p = [s for s in variants if 'iso_3p' in s]
            if len(var3p):
                var3p = int(var3p[0][7:])  # Position of iso_3p value
            else:
                var3p = 0

            var_add3p = [s for s in variants if 'iso_add3p' in s]
            if len(var_add3p):
                var_add3p = int(var_add3p[0][10:])  # Position of iso_add3p value
            else:
                var_add3p = 0
            var3p = var3p + var_add3p
            logger.debug("VCF::VAR_3p::%s" % var3p)
            var5p = [s for s in variants if 'iso_5p' in s]
            if len(var5p):
                var5p = int(var5p[0][7:])  # Position of iso_5p value
            else:
                var5p = 0  #
            logger.debug("VCF::VAR_5p::%s" % var5p)
            cigar = gff_fields['attrb']["Cigar"]
            # Obtaining all the variants from the cigar:
            if 1:
                (key_pos, key_var, vcf_ref, vcf_alt) = cigar_2_key(cigar, gff_fields['attrb']['Read'], ref_seq, vcf_pos,
                                                                   var5p, var3p, parent_ini_pos, parent_end_pos, hairpin)

                # Adding the variants to a dictionary and calculating all the fields of a vcf file format:
                if len(key_var) > 0:
                    for s in range(len(key_var)):
                        key_dict = vcf_chrom + '-' + str(key_pos[s]) + '-' + str(key_var[s])
                        raw_counts = gff_fields['attrb']['Expression']
                        raw_counts = [int(i) for i in raw_counts.split(',')]
                        nozero_counts = [int(i > 0) for i in raw_counts]  # counts for every sample if expr != 0.
                        if gtf_name in mirna_dict:  # Adding expression values to same mirnas
                            mirna_dict[gtf_name]['Z'] = [sum(x) for x in zip(mirna_dict[gtf_name]['Z'], raw_counts)]
                        else:
                            mirna_dict[gtf_name] = {}
                            mirna_dict[gtf_name]["Z"] = raw_counts
                        if key_dict in all_dict:
                            if all_dict[key_dict]["Type"] in ["A", "C", "T", "G"]:
                                all_dict[key_dict]['X'] = [sum(x) for x in zip(all_dict[key_dict]['X'], nozero_counts)]
                                all_dict[key_dict]['Y'] = [sum(x) for x in zip(all_dict[key_dict]['Y'], raw_counts)]
                        else:
                            key_list.append(key_dict)
                            all_dict[key_dict] = {}
                            all_dict[key_dict]["Chrom"] = vcf_chrom
                            all_dict[key_dict]["Position"] = key_pos[s]
                            all_dict[key_dict]["mirna"] = gtf_name
                            all_dict[key_dict]["Type"] = key_var[s]
                            if key_var[s][0] in ["A", "C", "T", "G"]:
                                n_SNP += 1
                                all_dict[key_dict]["SNP"] = True
                                all_dict[key_dict]["ID"] = gff_fields['attrb']['Name'] + '-SNP' + str(n_SNP)
                                all_dict[key_dict]['X'] = nozero_counts
                                all_dict[key_dict]['Y'] = raw_counts
                            else:
                                n_noSNP += 1
                                all_dict[key_dict]["SNP"] = False
                                all_dict[key_dict]["ID"] = gff_fields['attrb']['Name'] + '-nonSNP' + str(n_noSNP)
                            all_dict[key_dict]["Ref"] = vcf_ref[s]
                            all_dict[key_dict]["Alt"] = vcf_alt[s]
                            all_dict[key_dict]["Qual"] = "."
                            all_dict[key_dict]["Filter"] = gff_fields['attrb']['Filter']
                            all_dict[key_dict]["Info"] = "NS=" + str(len(sample_names))
            else:
                no_var += 1

    #  Writing the VCF file:
    for s in key_list:
        variant_line = ("\n%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s" %
                       (all_dict[s]["Chrom"], all_dict[s]["Position"], all_dict[s]["ID"],
                        all_dict[s]["Ref"], all_dict[s]["Alt"], all_dict[s]["Qual"],
                        all_dict[s]["Filter"], all_dict[s]["Info"]))
        if all_dict[s]["Type"] in ["A", "T", "C", "G"]:
            format_col = "TRC:TSC:TMC:GT"
            variant_line = variant_line + "\t" + format_col
            samples = ""
            for n in range(len(sample_names)):
                X = all_dict[s]["X"][n]
                Y = all_dict[s]["Y"][n]
                Z = mirna_dict[all_dict[s]["mirna"]]["Z"][n]
                # Calculating the genotype:
                if Y == 0:
                    GT = "0|0"
                elif Z == Y:
                    GT = "1|1"
                else:
                    GT = "1|0"
                samples = samples + "\t" + str(X) + ":" + str(Y) + ":" + str(Z) + ":" + GT
            variant_line = variant_line + samples
        else:
            format_col = ""
            variant_line = variant_line + format_col
        vcf_file.write(variant_line)
    vcf_file.close()



"""Read database information"""

from collections import defaultdict
import re

import mirtop.libs.logger as mylog

logger = mylog.getLogger(__name__)


def guess_database(args):
    """
    Guess database name from GFF file.

    Args:
        *gtf(str)*: file name with GFF miRNA genomic positions and
            header lines.

    Returns:
        *database(str)*: name of the database

    TODO: this needs to be generic to other databases.
    """
    if not hasattr(args, "database"):
        args.database = None
    return _guess_database_file(args.gtf, args.database)


def _guess_database_file(gff, database=None):
    if database:
        return database
    with open(gff) as in_handle:
        for line in in_handle:
            if not line.startswith("#"):
                break
            if line.find("miRBase") > -1:
                database = line[line.find("miRBase"):].strip().replace(" ", "")
            elif line.find("MirGeneDB") > -1:
                database = line[line.find("MirGeneDB"):].strip().replace(" ", "")
            elif line.find("microRNAs") > -1:
                database = line.strip().split()[1]
    if not database:
        logger.error("Database not found in --mirna %s. "
                     "Use --database argument to add a custom source." % gff)
        raise ValueError("Database not found in %s header" % gff)
    return database


def get_primary_transcript(database):
    """
    Get the ID to identify the primary transcript in the
      GTF file with the miRNA and precursor coordinates
      to be able to parse BAM files with genomic
      coordinates.
    """
    if database.find("miRBase") > -1:
        return "miRNA_primary_transcript"
    else:
        raise ValueError("Only miRBase is supported for this action.")


def read_gtf_to_mirna(gtf, database=None):
    """
    Load GTF file with precursor positions on genome.

    Args:
        *gtf(str)*: file name with GFF miRNA genomic positions and
            header lines.

    Returns:
        *db_mir(dict)*: dictionary with keys being mirnas and values
            genomic positions.
    """
    if not gtf:
        return gtf
    if not database:
        database = _guess_database_file(gtf)
    if database.find("miRBase") > -1:
        mapped = read_gtf_to_precursor_mirbase(gtf, format="genomic")
    elif database.find("MirGeneDB") > -1:
        mapped = read_gtf_to_precursor_mirgenedb(gtf, format="genomic")
    else:
        logger.info("Database different than miRBase or MirGeneDB")
        logger.info("If you get an error when loading,")
        logger.info("report it to https://github.com/miRTop/mirtop/issues")
        mapped = read_gtf_to_precursor_mirbase(gtf)
    return mapped


def read_gtf_chr2mirna(gtf):  # read from read_gtf_to_precursor
    """
    Load GTF file with precursor positions on genome.

    Args:
        *gtf(str)*: file name with GFF miRNA genomic positions and
            header lines.

    Returns:
        *db_mir(dict)*: dictionary with keys being chr and values
            mirna and genomic positions.
    """
    if not gtf:
        return gtf
    if _guess_database_file(gtf).find("miRBase") > -1:
        mapped = read_gtf_to_precursor_mirbase(gtf, format="chrom")
    elif _guess_database_file(gtf).find("MirGeneDB") > -1:
        mapped = read_gtf_to_precursor_mirgenedb(gtf, format="chrom")
    else:
        logger.info("Database different than miRBase or MirGeneDB")
        logger.info("If you get an error when loading,")
        logger.info("report it to https://github.com/miRTop/mirtop/issues")
        mapped = read_gtf_to_precursor_mirbase(gtf)
    return mapped


def read_gtf_chr2mirna2(gtf):  # to remove
    mir2hairpin = read_gtf_to_precursor(gtf)
    if not gtf:
        return gtf
    db = defaultdict(list)
    db_mir = defaultdict(list)
    id_dict = dict()
    with open(gtf) as in_handle:
        for line in in_handle:
            if line.startswith("#"):
                continue
            cols = line.strip().split("\t")
            logger.debug("MAP:: line:%s" % cols)
            name = [n.split("=")[1] for n in cols[-1].split(";")
                    if n.startswith("Name")]
            idname = [n.split("=")[1] for n in cols[-1].split(";")
                      if n.startswith("ID")]
            chrom, start, end, strand = cols[0], cols[3], cols[4], cols[6]
            logger.debug("MAP:: idname:%s" % idname)
            logger.debug("MAP:: name:%s" % name)
            id_dict[idname[0]] = name[0] # MIMA to sps-Y-X-5/3p
            if cols[2] == "miRNA_primary_transcript":
                db[idname[0]] = [chrom, int(start), int(end), strand]
            if cols[2] == "miRNA":
                parent = [n.split("=")[1] for n in cols[-1].split(";")
                          if n.startswith("Derives_from")]
                parent_name = id_dict[parent[0]]
                db_mir[chrom].append([name[0], start, end,
                                      strand,
                                      parent_name,
                                      mir2hairpin[parent_name][name[0]][0]])
                logger.debug("MAP:: mirna:%s" % name[0])
                logger.debug("MAP:: precursor:%s" % parent_name)
                logger.debug("MAP:: precursor pos %s" % mir2hairpin[parent_name][name[0]])
    return db_mir


def read_gtf_to_precursor(gtf, database=None):
    """
    Load GTF file with precursor positions on genome
    Return dict with key being precursor name and
    value a dict of mature miRNA with relative position
    to precursor.

    Args:
        *gtf(str)*: file name with GFF miRNA genomic positions and
            header lines.

    Returns:
        *map_dict(dict)*:

        >>> {'parent': {mirna: [start, end]}}
    """
    if not gtf:
        return gtf
    if not database:
        database = _guess_database_file(gtf)
    if database.find("miRBase") > -1:
        mapped = read_gtf_to_precursor_mirbase(gtf)
    elif database.find("MirGeneDB") > -1:
        mapped = read_gtf_to_precursor_mirgenedb(gtf)
    else:
        logger.info("Database different than miRBase or MirGeneDB")
        logger.info("If you get an error when loading,")
        logger.info("report it to https://github.com/miRTop/mirtop/issues")
        try:
            mapped = read_gtf_to_precursor_mirbase(gtf)
            return mapped
        except Exception as e:
            print(f"Failed to parse with Mirbase: {e}")
            try:
                mapped = read_gtf_to_precursor_mirgenedb(gtf)
                return mapped
            except Exception as e:
                print(f"Failed to parse with Mirgenedb: {e}")
                raise ValueError(f"There is no parser available for the database that you used: {database}")

    return mapped


def _parse_db_mir_chrom(db_mir, db, id_dict):
    mir2hairpin = _parse_db_mir(db_mir, db, id_dict)
    map_dict = defaultdict(list)
    for mir in db_mir:
        parent = db_mir[mir][4]
        precursor = db[id_dict[parent]]
        logger.debug("MAP::%s %s %s" % (
            id_dict[parent], precursor[1], precursor[2]))
        logger.debug("MAP::%s %s %s" % (mir, db_mir[mir][1], db_mir[mir][2]))
        if precursor[3] != db_mir[mir][3]:
            logger.warning("%s -> %s" % (id_dict[parent], mir))
            logger.warning("miRNA strand doesn't match with precursor strand:"
                           " %s - %s" % (db_mir[mir][3], precursor[3]))
            next
        if precursor[0] != db_mir[mir][0]:
            logger.warning("%s -> %s" % (id_dict[parent], mir))
            logger.warning("miRNA chr doesn't match with precursor chr:"
                           " %s - %s" % (db_mir[mir][0], precursor[0]))
            next
        map_dict[db_mir[mir][0]].append([mir[1], db_mir[mir][1], db_mir[mir][2],
                                         db_mir[mir][3],
                                         id_dict[parent],
                                         mir2hairpin[id_dict[parent]][mir[1]][0]])
    return map_dict


def _parse_db_mir(db_mir, db, id_dict):
    map_dict = defaultdict(dict)
    for mir in db_mir:
        parent = db_mir[mir][4]
        precursor = db[id_dict[parent]]
        logger.debug("MAP::%s %s %s" % (
            id_dict[parent], precursor[1], precursor[2]))
        logger.debug("MAP::%s %s %s" % (mir, db_mir[mir][1], db_mir[mir][2]))
        if precursor[3] != db_mir[mir][3]:
            logger.warning("%s -> %s" % (id_dict[parent], mir))
            logger.warning("miRNA strand doesn't match with precursor strand:"
                           " %s - %s" % (db_mir[mir][3], precursor[3]))
            next
        if precursor[0] != db_mir[mir][0]:
            logger.warning("%s -> %s" % (id_dict[parent], mir))
            logger.warning("miRNA chr doesn't match with precursor chr:"
                           " %s - %s" % (db_mir[mir][0], precursor[0]))
            next
        if precursor[3] == "+":
            start = db_mir[mir][1] - precursor[1]
            end = db_mir[mir][2] - precursor[1]
        if precursor[3] == "-":
            end = precursor[2] - db_mir[mir][1]
            start = precursor[2] - db_mir[mir][2]
        logger.debug("MAP:: final:%s %s %s" % (mir[1], start, end))
        map_dict[id_dict[parent]][mir[1]] = [start, end, db_mir[mir][3]]
    return map_dict


def _parse_db_mir_genomic(db_mir, db, id_dict):
    map_dict = defaultdict(dict)
    for mir in db_mir:
        parent = db_mir[mir][4]
        precursor = db[id_dict[parent]]
        logger.debug("MAP::%s %s %s" % (
            id_dict[parent], precursor[1], precursor[2]))
        logger.debug("MAP::%s %s %s" % (mir, db_mir[mir][1], db_mir[mir][2]))
        if precursor[3] != db_mir[mir][3]:
            logger.warning("%s -> %s" % (id_dict[parent], mir))
            logger.warning("miRNA strand doesn't match with precursor strand:"
                           " %s - %s" % (db_mir[mir][3], precursor[3]))
            next
        if precursor[0] != db_mir[mir][0]:
            logger.warning("%s -> %s" % (id_dict[parent], mir))
            logger.warning("miRNA chr doesn't match with precursor chr:"
                           " %s - %s" % (db_mir[mir][0], precursor[0]))
            next
        map_dict[mir[1]].update({id_dict[parent]: db_mir[mir]})
    return map_dict


def read_gtf_to_precursor_mirgenedb(gtf, format="precursor"):
    """
    Load GTF file with precursor positions on genome
    Return dict with key being precursor name and
    value a dict of mature miRNA with relative position
    to precursor. For MirGeneDB and similar GFF3 files.

    Args:
        *gtf(str)*: file name with GFF miRNA genomic positions and
            header lines.

    Returns:
        *map_dict(dict)*:

        >>> {'parent': {mirna: [start, end]}}
    """
    if not gtf:
        return gtf
    db = defaultdict(list)
    db_mir = defaultdict(list)
    id_dict = dict()
    pattern = r'(_3p\*?|_5p\*?)'
    with open(gtf) as in_handle:
        for line in in_handle:
            if line.startswith("#"):
                continue
            cols = line.strip().split("\t")
            idname = [n.split("=")[1] for n in cols[-1].split(";")
                      if n.startswith("ID")][0]
            name = idname
            chrom, start, end, strand = cols[0], cols[3], cols[4], cols[6]
            id_dict[idname] = name
            if cols[2] == "pre_miRNA":
                db[name] = [chrom, int(start), int(end), strand]
            if cols[2] == "miRNA":
                idname_mi = [n.split("=")[1] for n in cols[-1].split(";")
                             if n.startswith("ID")][0]
                # parent = "%s_pre" % idname_mi.replace("_3p.*", "").replace("_5p.*", "")
                parent = re.sub(pattern, '', idname_mi)
                parent = "%s_pre" % parent
                # import pdb; pdb.set_trace()
                db_mir[(parent, name)] = [chrom,
                                          int(start), int(end),
                                          strand, parent]
                logger.debug("MAP:: mirna:%s" % name)
                logger.debug("MAP:: pos %s" % db_mir[(parent, name)])
    if format == "genomic":
        return _parse_db_mir_genomic(db_mir, db, id_dict)
    elif format == "chrom":
        return _parse_db_mir_chrom(db_mir, db, id_dict)
    return _parse_db_mir(db_mir, db, id_dict)


def read_gtf_to_precursor_mirbase(gtf, format="precursor"):
    """
    Load GTF file with precursor positions on genome
    Return dict with key being precursor name and
    value a dict of mature miRNA with relative position
    to precursor. For miRBase and similar GFF3 files.

    Args:
        *gtf(str)*: file name with GFF miRNA genomic positions and
            header lines.

    Returns:
        *map_dict(dict)*:

        >>> {'parent': {mirna: [start, end]}}
    """
    if not gtf:
        return gtf
    db = defaultdict(list)
    db_mir = defaultdict(list)
    id_dict = dict()
    with open(gtf) as in_handle:
        for line in in_handle:
            if line.startswith("#"):
                continue
            cols = line.strip().split("\t")
            name = [n.split("=")[1] for n in cols[-1].split(";")
                    if n.startswith("Name")]
            idname = [n.split("=")[1] for n in cols[-1].split(";")
                      if n.startswith("ID")]
            chrom, start, end, strand = cols[0], cols[3], cols[4], cols[6]
            id_dict[idname[0]] = name[0]
            if cols[2] == "miRNA_primary_transcript":
                db[name[0]] = [chrom, int(start), int(end), strand]
            if cols[2] == "miRNA":
                parent = [n.split("=")[1] for n in cols[-1].split(";")
                          if n.startswith("Derives_from")]
                db_mir[(parent[0], name[0])] = [chrom,
                                                int(start), int(end),
                                                strand, parent[0]]
                logger.debug("MAP:: mirna:%s" % name[0])
                logger.debug("MAP:: pos %s" % db_mir[(parent[0], name[0])])
    if format == "genomic":
        return _parse_db_mir_genomic(db_mir, db, id_dict)
    elif format == "chrom":
        return _parse_db_mir_chrom(db_mir, db, id_dict)
    return _parse_db_mir(db_mir, db, id_dict)


def liftover_genomic_precursor(read, genome, hairpin, expected=None):
    # example 1 LESS (+) strand to miRNAX (+)1010 (->10)
    # genome miRNAX [start, end, strand] (+)1009
    # hairpin miRNAX [start, end] 9 (1 MORE = 8, 1 LESS = 10)
    # hairpin_start - (genome_start - read_start)
    # 9 - (1009 - 1010) = 9 - (-1) = 10
    if read['strand'] != genome['strand']:
        # Add warning
        return None
    logger.debug("MAPPER::liftover::genomic %s" % genome)
    logger.debug("MAPPER::liftover::hairpin %s" % hairpin)
    logger.debug("MAPPER::liftover::read %s" % read)
    if genome['strand'] == "+":
        lifted = hairpin["start"] - (genome["start"] - read["start"])
    # example 1 MORE: 1008
    # 9 - (1009 - 1008) = 9 - 1 = 8
    # example 1 MORE (-) strand: 1010
    # hairpin_start - (read_start - genome_start)
    # 9 - (1010 - 1009) = 9 - 1 = 8
    # example 1 LESS: 1008
    #  9 - (1008 - 1009) = 9 - (-1) = 10
    elif genome['strand'] == "-":
        lifted = hairpin["start"] - (read["end"] - genome["end"])
    logger.debug("MAPPER::liftover::lifted %s" % lifted)
    if expected and expected != lifted:
        raise ValueError("Bad liftover event.")
    return lifted



"""Main functions to simulates reads"""

import sys

from mirtop.libs.logger import initialize_logger
from mirtop.libs.parse import parse_cl
from mirtop.libs.simulator import simulate
from mirtop.gff import reader
from mirtop.gff.stats import stats
from mirtop.gff.compare import compare
from mirtop.gff.convert import convert_gff_counts
from mirtop.exporter import export
from mirtop.gff import validator
from mirtop.libs import spikeins
from mirtop.gff import update
from mirtop.sql import sql 
from mirtop.mirna import mapper
import mirtop.libs.logger as mylog

import time


def main(**kwargs):
    kwargs = parse_cl(sys.argv[1:])
    initialize_logger(kwargs['args'].out, kwargs['args'].debug,
                      kwargs['args'].print_debug)
    logger = mylog.getLogger(__name__)
    start = time.time()
    if not hasattr(kwargs["args"], "database"):
        if ("sql" not in kwargs and "stats" not in kwargs and "update" not in kwargs and "validate" not in kwargs):
            kwargs["args"].database = mapper.guess_database(kwargs["args"])

    if "gff" in kwargs:
        logger.info("Run annotation")
        reader(kwargs["args"])
    elif "stats" in kwargs:
        logger.info("Run stats.")
        stats(kwargs["args"])
    elif "compare" in kwargs:
        logger.info("Run compare.")
        compare(kwargs["args"])
    elif "simulator" in kwargs:
        logger.info("Run simulation")
        simulate(kwargs["args"])
    elif "counts" in kwargs:
        logger.info("Run convert of GFF to TSV containing expression")
        convert_gff_counts(kwargs["args"])
    elif "export" in kwargs:
        logger.info("Run export of GFF into other format.")
        export(kwargs["args"])
    elif "validator" in kwargs:
        logger.info("Run validator.")
        validator.check_multiple(kwargs["args"])
    elif "query" in kwargs:
        logger.info("Not yet ready: This will allow queries to GFF files.")
    elif "spikein" in kwargs:
        logger.info("Run spike-in tools")
        spikeins.convert(kwargs["args"])
    elif "update" in kwargs:
        logger.info("Run update tools")
        update.convert(kwargs["args"])
    elif "sql" in kwargs:
        logger.info("Run Convert GFF.")
        sql.sql_options(kwargs["args"])
    logger.info('It took %.3f minutes' % ((time.time()-start)/60))



""" Read seqbuster files"""

from collections import defaultdict

import mirtop.libs.logger as mylog
from mirtop.mirna.realign import isomir, hits
from mirtop.bam import filter
from mirtop.gff import body
from mirtop.mirna.annotate import annotate

logger = mylog.getLogger(__name__)


def header():
    """
    Custom header for seqbuster importer.

    Returns:
        *(str)*: seqbuster header string.
    """
    h = ("## CMD: seqbuster: http://seqcluster.readthedocs.io/mirna_annotation.html#mirna-isomir-annotation-with-java\n"
        "# iso_snp are not filtered yet. Use isomiRs R pacakge to correct for error sequencing\n")
    return h


def read_file(fn, args):
    """
    Read seqbuster file and convert to mirtop GFF format.

    Args:
        *fn(str)*: file name with seqbuster output information.

        *database(str)*: database name.

        *args(namedtuple)*: arguments from command line.
            See *mirtop.libs.parse.add_subparser_gff()*.

    Returns:
        *reads*: dictionary where keys are read_id and values are *mirtop.realign.hits*

    """
    precursors = args.precursors
    reads = defaultdict(hits)
    col_fix = 0
    with open(fn) as handle:
        header = handle.readline()
        if header.find("freq") < 0:
            col_fix = 1
        for line in handle:
            reads.update(_read_line(line, col_fix, precursors))
    logger.info("Hits: %s" % len(reads))
    return reads


def read_file_low_memory(fn, sample, args, out_handle):
    precursors = args.precursors
    reads = defaultdict(hits)
    col_fix = 0
    with open(fn) as handle:
        header = handle.readline()
        if header.find("freq") < 0:
            col_fix = 1
        for line in handle:
            reads = _read_line(line, col_fix, precursors)
            ann = annotate(reads, args.matures, args.precursors, quiet=True)
            gff_lines = body.create(ann, args.database, sample, args, quiet=True)
            body.write_body_on_handle(gff_lines, out_handle)


def _read_line(line, col_fix, precursors):
    reads = defaultdict(hits)
    cols = line.strip().split("\t")
    query_name = cols[1]
    query_sequence = cols[0]
    reference_start = int(cols[4-col_fix]) - 1
    seqbuster_iso = ":".join(cols[6-col_fix:10-col_fix])
    if query_sequence and query_sequence.find("N") > -1:
        return reads
    if query_name not in reads:
        reads[query_name].set_sequence(query_sequence)
        reads[query_name].counts = _get_freq(query_name)
    chrom = cols[13-col_fix]
    logger.debug("\nSEQBUSTER::NEW::query: {query_sequence}\n"
                 "  precursor: {chrom}\n"
                 "  name:  {query_name}\n"
                 "  start: {reference_start}\n"
                 "  iso: {seqbuster_iso}".format(**locals()))
    # logger.debug("SEQBUSTER:: cigar {cigar}".format(**locals()))
    iso = isomir()
    iso.align = line
    iso.set_pos(reference_start, len(reads[query_name].sequence))
    logger.debug("\nSEQBUSTER:: start %s end %s" % (iso.start, iso.end))
    if len(precursors[chrom]) < reference_start + len(reads[query_name].sequence):
        logger.debug("\nSEQBUSTER::len precursor" % len(precursors[chrom]))
        return reads
    iso.subs, iso.add, iso.cigar = filter.tune(reads[query_name].sequence,
                                               precursors[chrom],
                                               reference_start, None)
    logger.debug("\nSEQBUSTER::After tune start %s end %s" % (iso.start, iso.end))
    if len(iso.subs) < 6:
        logger.debug("\nSEQBUSTER::iso.subs %s - length %s" % (iso.subs, len(iso.subs)))
        reads[query_name].set_precursor(chrom, iso)
    return reads


def _get_freq(name):
    """
    Check if name read contains counts (_xNumber)
    """
    try:
        counts = int(name.split("_x")[1])
    except:
        return 0
    return counts



from mirtop.exporter import seqbuster, isomirs, fasta, vcf


def export(args):
    """
    GFF3 to others formats
    """
    if args.format == "seqbuster":
        seqbuster.convert(args)
    if args.format == "isomir":
        isomirs.convert(args)
    elif args.format == "fasta":
        fasta.convert(args)
    elif args.format == "vcf":
        vcf.convert(args)



"""GFF reader and creator helpers"""
from __future__ import print_function

from collections import defaultdict, OrderedDict
from mirtop.mirna.realign import get_mature_sequence, align_from_variants, \
    read_id, variant_to_5p, variant_to_3p, variant_to_add, \
    is_sequence, make_id
from mirtop.gff.header import read_samples
from mirtop.gff.classgff import feature

import mirtop.libs.logger as mylog
logger = mylog.getLogger(__name__)


def read(fn, args):
    """Read GTF/GFF file and load into annotate, chrom counts, sample, line"""
    samples = read_samples(fn)
    lines = defaultdict(dict)
    sep = " " if args.out_format == "gtf" else "="
    corrupted_uid = 0
    with open(fn) as inh:
        for line in inh:
            if line.startswith("#"):
                continue
            line = paste_columns(feature(line), sep=sep)
            gff = feature(line)
            cols = gff.columns
            attr = gff.attributes
            if attr['UID'] and not read_id(attr['UID']):
                corrupted_uid += 1
                continue
            if 'UID' not in attr:
                msg = "UID not found."
                if 'Read' not in attr:
                    if not is_sequence(attr['Read']):
                        msg = msg + " Sequence not valid in Read attribute."
                    else:
                        attr['UID'] = make_id(attr['Read'])
                if 'sequence' not in attr:
                    msg = msg + " Sequence not found in sequence attribute."
                    if not is_sequence(attr['sequence']):
                        msg = msg + " Sequence not valid in sequence attribute."
                    else:
                        attr['UID'] = make_id(attr['Read'])
            if 'UID' not in attr:
                logger.warning("Line is not a valid GFF3 line: %s" %
                               line.strip())
                logger.warning(msg)

            if cols['start'] not in lines[cols['chrom']]:
                lines[cols['chrom']][cols['start']] = []
            uid = "%s-%s-%s" % (attr['UID'],
                                attr['Variant'],
                                attr['Name'])
            if args.keep_name:
                uid = "%s-%s" % (uid, attr['Read'])
            lines[cols['chrom']][cols['start']].append(
                [uid,
                 cols['chrom'],
                 attr['Expression'].strip().split(","),
                 samples,
                 line.strip()])
    logger.info("Lines skipped due to corrupted UID: %s" % corrupted_uid)
    return lines


def write_body_on_handle(lines, out_handle):
    for m in lines:
        for s in sorted(lines[m].keys()):
            for hit in lines[m][s]:
                print(hit[4], file=out_handle)


def create(reads, database, sample, args, quiet=False):
    """Read https://github.com/miRTop/mirtop/issues/9"""
    sep = " " if args.out_format == "gtf" else "="
    seen = set()
    lines = defaultdict(defaultdict)
    seen_ann = {}
    filter_precursor = 0
    filter_score = 0
    n_hits = 0
    n_reads = 0
    n_seen = 0
    if args.add_extra:
        precursors = args.precursors
        matures = args.matures
    for (r, read) in reads.items():
        hits = set()
        [hits.add(mature.mirna) for mature in read.precursors.values()
            if mature.mirna]
        hits = len(hits)
        if len(read.precursors) > 0:
            n_reads += 1
        for (ps, iso) in read.precursors.items():
            p = list(ps)[0]
            if not iso.mirna:
                filter_precursor += 1
                continue
            if (r, iso.mirna) not in seen:
                seen.add((r, iso.mirna))
                chrom = p
                seq = reads[r].sequence
                seq_name = seq if not args.keep_name else r
                if iso.get_score(len(seq)) < 1:
                    filter_score += 1
                    continue
                if iso.subs:
                    iso.subs = [] if "N" in iso.subs[0] else iso.subs
                idseq = reads[r].idseq
                source = "ref_miRNA" if not iso.is_iso() else "isomiR"
                strand = iso.strand
                start, end = iso.start, iso.end
                score = iso.map_score
                mirName = iso.mirna
                preName = p
                Variant = iso.formatGFF()
                Cigar = iso.cigar
                counts = read.counts
                Filter = iso.filter
                annotation = "%s.%s.%s" % (chrom, idseq, seq_name)
                # TODO:  This need to be moved to use the feature class
                # It needs a dict with all variable in keys
                fields = {'seq_name': seq_name, 'idseq': idseq,
                          'name': mirName, 'parent': preName,
                          'variant': Variant, 'cigar': Cigar,
                          'counts': counts, 'filter': Filter,
                          'hits': hits, 'chrom': chrom, 
                          'start': start, 'end': end,
                          'database': database, 'source': source,
                          'score': score, 'strand': strand}
                line = feature(fields).line
                logger.debug("GFF::%s" % line)
                if args.add_extra:
                    extra = variant_with_nt(line, precursors, matures)
                    line = "%s Changes %s;" % (line, extra)

                if annotation in seen_ann and seq.find("N") < 0 and (
                        seen_ann[annotation].split("\t")[0].find("N") < 0):
                    logger.warning(
                        "Same isomir %s from different sequence:"
                        " \n%s and \n%s" % (annotation, line,
                                            seen_ann[annotation]))
                seen_ann[annotation] = line
                logger.debug("GFF::external %s" % iso.external)
                if start not in lines[chrom]:
                    lines[chrom][start] = []
                lines[chrom][start].append([annotation, chrom,
                                            counts, sample, line])
                logger.debug("GFF::%s" % line)
                n_hits += 1
            else:
                n_seen += 1
    if not quiet:
        logger.info("GFF miRNAs: %s" % len(lines))
        logger.info("GFF hits %s by %s reads" % (n_hits, n_reads))
        logger.info("Filtered by being duplicated: %s" % n_seen)
        logger.info("Filtered by being outside miRNA positions:"
                    " %s" % filter_precursor)
        logger.info("Filtered by being low score: %s" % filter_score)
    return lines


def lift_to_genome(line, mapper):
    """
    Function to get a class of type feature from classgff.py
      and map the precursors coordinates to the genomic coordinates
    
    Args:
        *line(str)*: string GFF line.
        *mapper(dict)*: dict with mirna-precursor-genomic coordinas from
          mirna.mapper.read_gtf_to_mirna function.

    Returns:
        *(line)*: string with GFF line with updated chr, star, end, strand
 
    """
    features = feature(line)
    features.attributes["Variant"]
    chr, start, end, strand, id = mapper[features.attributes["Name"]][features.attributes["Parent"]]
    logger.debug("LIFT2GENOME:: %s of %s found in %s(%s) " % (features.attributes["Name"],
                                                            features.attributes["Parent"],
                                                            chr, strand))
    nstart = start
    nend = end
    variants = read_variant(features.attributes["Variant"])
    logger.debug("LIFT2GENOME:: variants %s " % (features.attributes["Variant"]))
    if 'iso_5p' in variants:
        if strand == "+":
            nstart = start + variants['iso_5p']
        else:
            nend = end - variants['iso_5p']
    if 'iso_3p' in variants:
        if strand == "+":
            nend = end + variants['iso_3p']
        else:
            nstart = start - variants['iso_3p']
    if 'iso_add3p' in variants:
        if strand == "+":
            nend = nend + variants['iso_add3p']
        else:
            nstart = nstart - variants['iso_add3p']
    logger.debug("LIFT2GENOME:: start %s to %s |  end %s to %s " % (start, nstart, end, nend))
    features.columns['chrom'] = chr
    features.columns['start'] = str(start)
    features.columns['end'] = str(end)
    features.columns['strand'] = strand
    
    return features.paste_columns()


def create_line(read, name, database, args):
    sep = " " if args.out_format == "gtf" else "="

    if args.add_extra:
        precursors = args.precursors
        matures = args.matures

    for (ps, iso) in read.precursors.items():
        p = list(ps)[0]
        if not iso.mirna:
            continue
        chrom = p
        seq = read.sequence
        seq_name = seq if not args.keep_name else name
        if iso.get_score(len(seq)) < 1:
            continue
        if iso.subs:
            iso.subs = [] if "N" in iso.subs[0] else iso.subs
        idseq = read.idseq
        source = "ref_miRNA" if not iso.is_iso() else "isomiR"
        strand = iso.strand
        start, end = iso.start, iso.end
        score = iso.map_score
        mirName = iso.mirna
        preName = p
        Variant = iso.formatGFF()
        Cigar = iso.cigar
        counts = read.counts
        Filter = iso.filter
        annotation = "%s.%s.%s" % (chrom, idseq, seq_name)
        # This get correctly formated with paste_columns below
        attrb = ("Read {seq_name};UID {idseq};Name {mirName};"
                 "Parent {preName};"
                 "Variant {Variant};Cigar {Cigar};"
                 "Expression {counts};"
                 "Filter {Filter};").format(**locals())
        line = ("{chrom}\t{database}\t{source}\t{start}\t{end}"
                "\t{score}\t{strand}\t.\t{attrb}").format(**locals())
        logger.debug("GFF::%s" % line)
        if args.add_extra:
            extra = variant_with_nt(line, precursors, matures)
            line = "%s Changes %s;" % (line, extra)

        line = feature(line).paste_columns(sep)
        return line


def guess_format(line):
    return "=" if line.find("Name=") > -1 else " "


def paste_columns(line, sep=" "):
    """
    Create GFF/GTF line from read_gff_line
    """
    cols = line.columns
    attr = line.attributes
    attr_paste = "; ".join(
        "%s%s%s" % (a, sep, attr[a]) for a in attr)
    return "\t".join([cols['chrom'], cols['source'], cols['type'],
                      cols['start'], cols['end'], cols['score'],
                      cols['strand'], cols['ext'], attr_paste])


def read_variant(attrb, sep=" "):
    """
    Read string in variants attribute.

    Args:
        *attrb(str)*: string in Variant attribute.

    Returns:
        *(gff_dict)*: dictionary with:
            >>> {'iso_3p': -3, ...}
    """
    gff_dict = OrderedDict()
    logger.debug("variant: %s" % attrb)
    for gff_item in attrb.strip().split(","):
        item_pair = gff_item.strip().split(":")
        if len(item_pair) > 1:
            gff_dict[item_pair[0].strip()] = int(item_pair[1].strip())
        else:
            gff_dict[item_pair[0].strip()] = True
    logger.debug("VARIANTS::Keys found: %s" % gff_dict.keys())
    logger.debug("VARIANTS::Values found: %s" % gff_dict.values())
    return gff_dict


def read_attributes(gff_line, sep=" "):
    gff_line = gff_line.strip().split("\t")[8]
    gff_dict = OrderedDict()
    for gff_item in gff_line.strip().split(";"):
        item_pair = gff_item.strip().split(sep)
        if len(item_pair) > 1:
            gff_dict[item_pair[0].strip()] = item_pair[1].strip()
    return gff_dict


def read_gff_line(line):
    """
    Read GFF/GTF line and return dictionary with fields
    """
    if line.startswith("#"):
        return line
    cols = line.strip().split("\t")
    sep = guess_format(line)
    if len(cols) != 9:
        raise ValueError("Line has less than 9 elements: %s" % line)
    fields = {'chrom': cols[0],
              'source': cols[1],
              'type': cols[2],
              'start': cols[3],
              'end': cols[4],
              'score': cols[5],
              'strand': cols[6],
              'ext': cols[7],
              'attrb': read_attributes(line, sep)}
    return fields


def variant_with_nt(line, precursors, matures):
    """
    Return nucleotides changes for each variant type
    using Variant attribute, precursor sequences and
    mature position.
    """
    gff = feature(line)
    attr = gff.attributes
    read = read_id(attr["UID"])
    attr["Parent"] = attr["Parent"].split(",")[0]
    if attr["Parent"] not in matures:
        logger.warning("Parent miRNA not found in database %s" % attr["Parent"])
        return ""
    if attr["Name"] not in matures[attr["Parent"]]:
        logger.warning("miRNA not found in database %s" % attr["Name"])
        return ""

    logger.debug("GFF::BODY::precursors %s" % precursors[attr["Parent"]])
    logger.debug("GFF:BODY::mature %s" % matures[attr["Parent"]][attr["Name"]])

    t5 = variant_to_5p(precursors[attr["Parent"]],
                       matures[attr["Parent"]][attr["Name"]],
                       attr["Variant"])
    t3 = variant_to_3p(precursors[attr["Parent"]],
                       matures[attr["Parent"]][attr["Name"]],
                       attr["Variant"])
    add = variant_to_add(read,
                         attr["Variant"])
    mature_sequence = get_mature_sequence(
        precursors[attr["Parent"]],
        matures[attr["Parent"]][attr["Name"]],
        nt=8)
    logger.debug("GFF::BODY::mature_sequence %s" % mature_sequence)
    mm = align_from_variants(read,
                             mature_sequence,
                             attr["Variant"])
    if mm == "Invalid":
        return mm
    if len(mm) > 0:
        mm = "".join(["".join([str(v) for v in m]) for m in mm])
    else:
        mm = "0"
    return "iso_5p:%s,iso_3p:%s,iso_add3p:%s,iso_snv:%s" % (t5, t3, add, mm)



from collections import defaultdict
from mirtop.mirna.realign import hits, cigar_correction, make_cigar, align
import mirtop.libs.logger as mylog

logger = mylog.getLogger(__name__)


def tune(seq, precursor, start, cigar):
    """
    The actual fn that will realign the sequence to find the nt changes
    at 5', 3' sequence and nt variations.

    Args:
        *seq (str)*: sequence of the read.

        *precursor (str)*: sequence of the precursor.

        *start (int)*: start position of sequence on the precursor, +1.

        *cigar (str)*: similar to SAM CIGAR attribute.

    Returns:

        *list* with:

            subs (list): substitutions

            add (list): nt added to the end

            cigar (str): updated cigar
    """
    
    end = len(seq)
    if start < 0:
        end = end + start
        start = 0
    if cigar:
        seq, mature = cigar_correction(cigar, seq, precursor[start:])
    else:
        seq, mature, score, p, size = align(seq, precursor[start:start + end])
        cigar = make_cigar(seq, mature)
    if seq.startswith("-"):
        seq = seq[1:]
    if seq.endswith("-"):
        seq = seq[:-1]
    logger.debug("TUNE:: %s %s %s" % (cigar, seq, mature))
    subs, add = [], []

    if seq == mature:
        logger.debug("TUNE:: %s %s" % (subs, add))
        return subs, "".join(add), make_cigar(seq, mature)
    
    error = set()
    for pos in range(0, len(seq)):
        if seq[pos] != mature[pos]:
            error.add(pos)

    prob = 0
    add_position = []
    positions_to_look = [*range(len(seq) - 1, len(seq) - 6, -1)]
    is_overlapped = len([value for value in error if value in positions_to_look])
    if is_overlapped:
        for e in positions_to_look:
            if e in error:
                prob = 1
            if prob == 1:
                add.append(seq[e])
                add_position.append(e)
            if e not in error and prob == 0 and seq[e] in ["A", "T"]:
                add.append(seq[e])
                add_position.append(e)
                continue
            if e not in error:
                if add:
                    add.pop()
                    add_position.pop()
                if prob == 0:
                    add = []
                    add_position = []
                break

    for e in error:
        if e not in add_position:
            subs.append([e, seq[e], mature[e]])
    if not error:
        add = []
    
    logger.debug("TUNE:: %s %s" % (subs, add))
    
    return subs, "".join(add), make_cigar(seq, mature)


def clean_hits(reads):
    """
    Select only best matches from a list of hits from the same read.

    Args:
        *reads*: dictionary as:

        >>> {'read_id': mirtop.realign.hits, ...}

    Returns:

        *reads*: same than input but with best hits only.
    """
    new_reads = defaultdict(hits)
    for r in reads:
        world = {}
        sc = 0
        for p in reads[r].precursors:
            world[p] = reads[r].precursors[p].get_score(len(reads[r].sequence))
            if sc < world[p]:
                sc = world[p]
        new_reads[r] = reads[r]
        for p in world:
            logger.debug("CLEAN::score %s %s %s" % (r, p, world[p]))
            if sc != world[p]:
                logger.debug("CLEAN::remove %s %s %s" % (r, p, world[p]))
                new_reads[r].remove_precursor(p)

    return new_reads



from __future__ import print_function

import argparse
import os
import random
from collections import defaultdict

import pysam

import mirtop.libs.logger as mylog
import mirtop.libs.do as runner

parser = argparse.ArgumentParser()
parser.add_argument("--fa",
                    help="File with mature sequences.", required=True)
parser.add_argument("-o", "--out", default="spikeins.fa",
                    help="Name used for output files.")
parser.add_argument("--seed", help="set up seed for reproducibility.",
                    default=42)
parser.add_argument("--max_size", help="maximum size allowed in the final output.",
                    default=25)

args = parser.parse_args()
random.seed(args.seed)

def _sam_to_bam(bam_fn):
    bam_out = "%s.bam" % os.path.splitext(bam_fn)[0]
    cmd = "samtools view -Sbh {bam_fn} -o {bam_out}"
    runner.run(cmd.format(**locals()))
    return bam_fn


def _bam_sort(bam_fn):
    bam_sort_by_n = os.path.splitext(bam_fn)[0] + "_sort.bam"
    runner.run(("samtools sort -n -o {bam_sort_by_n} {bam_fn}").format(
               **locals()))
    return bam_sort_by_n

def _read_fasta(fa):
    source = dict()
    with open(fa) as inh:
        for line in inh:
            if line.startswith(">"):
                name = line.strip().split()[0].replace(">", "")
            else:
                source.update({name: line.strip()})
    return source


def _write_fasta(sequences, filename, max=25):
    with open(filename, 'w') as outh:
        for name in sequences:
            if sequences[name]:
                if len(sequences[name]) < max:
                    print(">%s\n%s" % (name, sequences[name]), file=outh)
    return filename


def _parse_hits(sam, source):
    uniques = defaultdict(list)
 #   bam_fn = _sam_to_bam(sam)
 #   bam_fn = _bam_sort(bam_fn)
 # read sequences and score hits (ignore same sequence)
    handle = pysam.Samfile(sam, "rb")
    for line in handle:
        reference = handle.getrname(line.reference_id)
        name = line.query_name
        # sequence = line.query_sequence if not line.is_reverse else reverse_complement(line.query_sequence)
        if reference == name:
            continue
        # print([reference, name, line.get_tag("NM")])
        distance = line.get_tag("NM")
        uniques[name].append(distance)
        uniques[reference].append(distance)
    # read parsed data and keep the ones with score > 10 edit distance
    for name in uniques:
        if min(uniques[name]) < 5:
            if name in source:
                source[name] = None
    return source

# Map all vs all with razers3
source = _read_fasta(args.fa)
sam = os.path.join(os.path.dirname(args.out), "modified.bam")
runner.run(("razers3 -dr 5 -i 75 -rr 80 -f -so 1 -o {output} {target} {query}").format(output=sam, target=args.fa, query=args.fa))
uniques = _parse_hits(sam, source)

# Write uniques to fasta
_write_fasta(uniques, args.out, args.max_size)



"""mirGFF3 proxy converter"""
from __future__ import print_function

import os.path as op

from mirtop.mirna import fasta, mapper
from mirtop.bam.bam import low_memory_bam, low_memory_genomic_bam
from mirtop.importer import seqbuster
from mirtop.gff import header
import mirtop.libs.logger as mylog
logger = mylog.getLogger(__name__)


def reader(args):
    """
    Realign BAM hits to miRBase to get better accuracy and annotation
    """
    samples = []
    database = mapper.guess_database(args)
    args.database = database
    precursors = fasta.read_precursor(args.hairpin, args.sps)
    args.precursors = precursors
    matures = mapper.read_gtf_to_precursor(args.gtf, args.database)
    args.matures = matures
    # TODO check numbers of miRNA and precursors read
    # TODO print message if numbers mismatch
    if args.keep_name and len(args.files) > 1:
        logger.warning("--keep-name when running multiple samples\n"
                       "can generate wrong results if the\n"
                       "name read is different across sample\n"
                       "for the same sequence.")
    for fn in args.files:
        fn = op.normpath(fn)
        if args.format != "gff":
            sample = op.splitext(op.basename(fn))[0]
            samples.append(sample)
            fn_out = op.join(args.out, sample + ".%s" % args.out_format)
        h = header.create([sample], args.database, "")
        out_handle = open(fn_out, 'w')
        print(h, file=out_handle)
        if args.format == "BAM":
            if args.genomic:
                low_memory_genomic_bam(fn, sample, out_handle, args)
            else:
                low_memory_bam(fn, sample, out_handle, args)
        elif args.format == "seqbuster":
            seqbuster.read_file_low_memory(fn, sample, args, out_handle)
        else:
            raise ValueError("%s not supported for low memory" % args.format)
        out_handle.close()



import matplotlib.pyplot as plt

def makePlots(tsvFileN, pdfFileN, show):
    #Reading file
    with open(tsvFileN, "r") as ins:
        lines = []
        for line in ins:
            lines.append(line.split('\t'))

    #Calculating maximum number of plots
    maxPlots = 0
    for x in range(1, len(lines)):
        idc = int(lines[x][0]) + 1
        if maxPlots < idc:
            maxPlots = idc

    # Set up the matplotlib figure
    cols = 3
    rows = maxPlots/3
    plt.subplots(3, rows, figsize=(8, 6), sharex=True)

    #Creating array to store the values
    array = []

    for i in range(0, maxPlots):
        arr2 = []
        for j in range(0, 5):
            arr2.append(0)
        array.append(arr2)

    #Filling the array with the values
    for x in range(1, len(lines)):
        idc = int(lines[x][0])
        cnt = int(lines[x][3])
        typ =     lines[x][4].strip()
        nam =     lines[x][1].strip()

        pos = 3;
        if typ == 'synthetic':
            pos = 2
        if typ == 'bcbio':
            pos = 0
        if typ == 'mirge':
            pos = 1
	
        array[idc][pos] = cnt
        array[idc][4]   = nam

    #Plotting the graphs
    plt.figure(1)
    #plt.xlabel('tool')
    #plt.ylabel('Counts')

    p = []
    p.append(array[0][0])
    p.append(array[0][1])
    p.append(array[0][2])
    n = array[0][4]

    for i in range(0, maxPlots):
        del(p[2])
        del(p[1])
        del(p[0])
        p.append(array[i][0])
        p.append(array[i][1])
        p.append(array[i][2])
        n = array[i][4]

        pcd = rows * 100 + cols * 10 + 1 + i
        plt.subplot(pcd)
    
        ax = plt.gca()
        ax.set_facecolor('lightgray')
    
        plt.xticks([1,2,3], ('bcbio', 'mirge', 'synthetic'))
        plt.yticks([0,10,20,30,40,50])
        plt.tick_params(axis='both', which='major', labelsize=8)
        plt.tick_params(axis='both', which='minor', labelsize=8)
    
        plt.bar([1,2,3], p, color='gray')
        plt.title(n)
    
        for i, v in enumerate(p):
            plt.text(i+0.9, 0, str(v), color='black', fontsize='8', fontweight='bold')

    plt.subplots_adjust(top=0.92, bottom=0.10, left=0.10, right=0.95, hspace=0.50, wspace=0.35)
    plt.savefig(pdfFileN, format="pdf")
    if show == 1:
        plt.show()


makePlots("../data/examples/plot/example_count.tsv", "kk.pdf", 1)



import os
import sys
from collections import defaultdict

from argparse import ArgumentParser

def _read_pri(fn):
    pri = dict()
    with open(fn) as inh:
        for line in inh:
            if line.startswith(">") & line.strip().endswith("pri"):
                name = line.strip()[1:-4]
            else:
                pri[name] = line.strip()
    return pri

def _read_bed(fn):
    bed = defaultdict(dict)
    with open(fn) as inh:
        for line in inh:
            cols = line.strip().split("\t")
            if cols[3].find("pri") > 0:
                continue
            if cols[3].find("loop") > 0:
                continue
            if cols[3].find("seed") > 0:
                continue
            if cols[3].find("motif") > 0:
                continue
            if cols[3].find("co") > 0:
                continue
            bed[cols[3].split("_")[0]].update({cols[3]: [int(cols[1]), int(cols[2]), cols[5]]})
    return bed


def _download(url, outfn):
    if os.path.isfile(outfn):
        return outfn
    os.system('wget -O %s %s' % (outfn, url))
    return outfn

if __name__ == "__main__":
    parser = ArgumentParser(description="Prepare files from mirGeneDB to be used with seqbuster")
    parser.add_argument("--bed", help="bed file with position of all sequence", required=1)
    parser.add_argument("--precursor30", help="file or url with fasta of precursor + 30 nt", required=1)
    args = parser.parse_args()

    sps = os.path.basename(args.precursor30).split("-")[0]

    if os.path.isfile(args.bed):
        fnbed = args.bed
    else:
        fnbed = _download(args.bed, "%s.bed" % sps)
    if os.path.isfile(args.precursor30):
        fnfa = args.precursor30
    else:
        fnfa = _download(args.precursor30, "%s.fa" % sps)
    fa = _read_pri(fnfa)
    bed = _read_bed(fnbed)
    OUT = open("%s.miRNA.str" % sps, 'w')
    OUTP = open("%s.hairpin.fa" % sps, 'w')
    for mir in fa:
        if mir in bed:
            precursor = bed[mir][mir + "_pre"]
            print(precursor)
            mir5p = ""
            mir3p = ""
            for mature in bed[mir]:
                info = bed[mir][mature]
                # print info
                if mature.endswith("pre"):
                    continue
                if precursor[2] == "-":
                    start = int(precursor[1]) - int(info[1]) + 31
                    end = int(precursor[1]) - int(info[0]) + 30
                else:
                    start = int(info[0]) - int(precursor[0]) + 31
                    end = int(info[1]) - int(precursor[0]) + 30
                    # print [mature, start, end, fa[mir][start:end]]
                if mature.find("5p") > 0:
                    mir5p = "[%s:%s-%s]" % (mature, start, end)
                if mature.find("3p") > 0:
                    mir3p = "[%s:%s-%s]" % (mature, start, end)

            print(">%s (X) %s %s" % (mir, mir5p, mir3p), file=OUT)
            print(">%s\n%s" % (mir, fa[mir]),file=OUTP)
OUT.close()
OUTP.close()



# import os
import argparse
import sqlite3
import re
from datetime import datetime
import time
import os.path as op

now = datetime.now()
# dd/mm/YY H:M:S
d2 = now.strftime("%B %d, %Y %H:%M:%S")


def create_table(conn, sample_names):
    c = conn.cursor()
    data_columns = ['seqID text', 'source_file text', 'type text', 'start real',
    'end real', 'score text', 'strand text', 'phase text', 'UID text', 'Read text', 'Name text', 'Parent text', 'Variant text',
    'iso_5p real', 'iso_3p real', 'iso_add3p real', 'iso_add5p real', 'iso_snv real', 'iso_snv_seed real', 'iso_snv_central real', 'iso_snv_central_offset real',
    'iso_snv_central_supp real', 'source text', 'cigar text', 'hits real', 'alias text', 'genomic_pos text', 'filter text',
    'seed_fam text']
    complete_headers = data_columns + sample_names
    q = "CREATE TABLE IF NOT EXISTS data_sets(%s)" % ", ".join(complete_headers)
    c.execute(q)
    conn.commit()


def gff_insert_values(conn, complete_list):
    try:
        conn.execute('INSERT INTO data_sets VALUES(' + ','.join("?" * len(complete_list)) + ')', complete_list)
        conn.commit()
    except sqlite3.OperationalError as e:
        print()
        print("ERROR:")
        print("sqlite3.OperationalError: {0}".format(e))
        print("Help: Make sure to delete any existing database with tables of different schema")
        exit()


# print("date and time =", d2)

def insert_sql(args):
    if args.db:
        out_file = op.join(args.out, args.db)
        conn = sqlite3.connect(out_file)
        c = conn.cursor()
    else:
        out_file = op.join(args.out, "mirtop.db")
        conn = sqlite3.connect(out_file)
        c = conn.cursor()

    with open(args.gff, 'r') as f:
        version = source = data_sets = tools = commands_exec = filter_tags = citation = num_records = ""
        cnt = 0
        for text in f:
            # HEADER INFORMATION
            if re.search("^## .* VERSION", text):  # (R)
                version = (text.strip().split(' ')[-1])
            elif re.search("^## source-ontology", text):  # (R)
                source = (text.strip().split(' ')[-1])
            elif re.search("^## COLDATA", text):  # (R)
                data_sets = (
                    text.strip().split(' ')[-1])  # Might contain more than one data set
                sample_names = data_sets.split(',')
                sample_names = [w.replace('-', '_') for w in sample_names]
                string_text = "text"
                output_sample_names = ["{} {}".format(i, string_text) for i in sample_names]
                create_table(conn, output_sample_names)
            elif re.search("^## TOOLS", text):  # (R)
                tools = (text.strip().split(' ')[-1])
            elif re.search("^## CMD", text):  # (O)
                commands_exec = (text.strip().split(' ')[-1])
            elif re.search("^## FILTER", text):  # (O)
                filter_tags = (text.strip().split(' ')[-1])
            elif re.search("^## REFERENCE", text):  # (O)
                citation = (text.strip().split(' ')[-1])
            # BODY - INFORMATION
            elif not re.search("^#", text):
                cnt += 1
                lines = text.strip().split('\t')
                if '=' in lines[-1]:
                    lines_info_array = lines[-1].replace("=", " ")
                else:
                    lines_info_array = lines[-1]

                info = lines_info_array.split('; ')
                info_dict = dict()
                for elements in info:
                    (k, v) = elements.split(' ')
                    info_dict.update([(k, v)])
                    if 'Variant' in k and ":" in v:
                        value_list = v.split(',')
                        for iso_vars in value_list:
                            if ":" in iso_vars:
                                (sub_k, sub_v) = iso_vars.split(':')
                                info_dict.update([(str(sub_k), str(sub_v))])
                            else:
                                ### Exception for miRge format START 
                                if iso_vars == "iso_snp":
                                    iso_vars = "iso_snv"
                                elif iso_vars == "iso_add": 
                                    iso_vars = "iso_add3p"
                                ### Exception for miRge format END
                                info_dict['iso_snv'] = "1" if iso_vars == 'iso_snv' else 0
                                info_dict['iso_snv_seed'] = "1" if iso_vars == 'iso_snv_seed' else 0
                                info_dict['iso_snv_central'] = "1" if iso_vars == 'iso_snv_central' else 0
                                info_dict['iso_snv_central_offset'] = "1" if iso_vars == 'iso_snv_central_offset' else 0
                                info_dict['iso_snv_central_supp'] = "1" if iso_vars == 'iso_snv_central_supp' else 0
                                 
                prefix_list = [lines[0], lines[1], lines[2], lines[3], lines[4], lines[5], lines[6], lines[7],
                               info_dict.get('UID'), info_dict.get('Read'), info_dict.get('Name'),
                               info_dict.get('Parent'),
                               info_dict.get('Variant'), 
                               str(info_dict.setdefault('iso_5p', None)),
                               str(info_dict.setdefault('iso_3p', None)), 
                               str(info_dict.setdefault('iso_add3p', None)),
                               str(info_dict.setdefault('iso_add5p', None)),
                               str(info_dict.setdefault('iso_snv', "0")), 
                               str(info_dict.setdefault('iso_snv_seed', "0")), 
                               str(info_dict.setdefault('iso_snv_central', "0")), 
                               str(info_dict.setdefault('iso_snv_central_offset', "0")), 
                               str(info_dict.setdefault('iso_snv_central_supp', "0")), 
                               source,
                               info_dict.setdefault('Cigar', None),
                               info_dict.setdefault('Hits', None), info_dict.setdefault('Alias', None),
                               info_dict.setdefault('Genomic', None),
                               info_dict.setdefault('Filter', None), info_dict.setdefault('Seed_fam', None)]
                expression_list = info_dict.get('Expression').split(',')
                complete_list = prefix_list + expression_list
                gff_insert_values(conn, complete_list)

        c.execute('''CREATE TABLE IF NOT EXISTS summary(version text, source text, data_sets text, tools text,
         commands_exec text, filter_tags text, citation text, records real, date_stamp text)''')
        c.execute("INSERT INTO summary(version, source, data_sets, tools, commands_exec, filter_tags, citation, "
                  "records, date_stamp) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
            (version, source, data_sets, tools, commands_exec, filter_tags, citation, cnt, d2))

        # info_dict.setdefault('Sex', None)

        conn.commit()


def query_sql(args):
    #print("Function query is being implemented, will be updated soon!!!")
    #print(args)
    if args.db:
        out_file = op.join(args.db)
        conn = sqlite3.connect(out_file)
        c = conn.cursor()
    else:
        out_file = op.join("mirtop.db")
        conn = sqlite3.connect(out_file)
        c = conn.cursor()

    #c.execute("SELECT name FROM sqlite_master WHERE type = 'table';")
    #record = c.fetchall()
    #print(args.db)
    if args.expr == "show-tables":
        show_tables(conn)
    if args.expr == "show-schema":
        if args.table:
            show_schema(conn, args.table)
        else:
            print("Error: Require table name")
            print("Usage: mirtop sql --query --db <input_database> -e show-schema -t <table_name>")
    if args.expr == "show-columns":
        show_columns(conn, args)
    if args.expr == "describe-gff":
        describe_gff_info(conn, args)
    if args.expr == "isomirs-per-mirna":
        if args.miRNA:
            stats_isomiR_per_miRNA(conn, args.miRNA, args)
        else:
            print("Error: Require miRNA name")
            print("Usage: mirtop sql --query --db <input_database> -e isomirs-per-mirna -miR <miRNA>")
    if args.expr == "select":
        select_query(conn, args)
    pass


def show_tables(connection):
    print(" +" + 25 * "-" + " +")
    print(' | Tables                   |')
    print(" +" + 25 * "-" + " +")
    for (tableName,) in connection.execute(
            """
        select NAME from SQLITE_MASTER where TYPE='table' order by NAME;
        """
    ):
        tn_name = len(tableName)
        req_format_space = 25 - tn_name
        print(" | {tn}{format_space}|".format(
            tn=tableName,
            format_space=req_format_space * " "
        ))  # Table name (for each table)
    print(" +" + 25 * "-" + " +")


def show_schema(connection, table_name):
    for (tableName,) in connection.execute(
            """
        select NAME from SQLITE_MASTER where TYPE='table' order by NAME;
        """
    ):
        #  print("{}:".format(tableName))  # Table name (for each table)
        if tableName == table_name:
            print(" +" + 57 * "-" + "+")
            print(' | Sl | Field                         | Type | NULL | Key  |')
            print(" +" + 57 * "-" + "+")
            for (
                    columnID, columnName, columnType,
                    columnNotNull, columnDefault, columnPK,
            ) in connection.execute("pragma table_info('{}');".format(tableName)):
                name_size = len(columnName)
                columnID_size = len(str(columnID))
                required_len = 30 - name_size
                req_col_size = 2 - columnID_size

                print(" | {colSpace}{id} | {name}{space}| {type} | {null} | {pk} |".format(
                    colSpace=req_col_size * " ",
                    id=columnID,
                    name=columnName,
                    space=required_len * " ",
                    type=columnType if columnType else "NULL",
                    null=" not null" if columnNotNull else " NO ",
                    default=" [{}]".format(columnDefault) if columnDefault else "NULL",
                    pk=" *{}".format(columnPK) if columnPK else "    ",
                ))
            print(" +" + 57 * "-" + "+")


def show_columns(connection, args):
    cur = connection.cursor()
    query="SELECT * FROM data_sets LIMIT 2"
    cur.execute(query)
    rows = cur.fetchall()
    col_name_list = [tuple[0] for tuple in cur.description]
    sl_no=1
    print("\nSerial\tColumn names")
    for col in col_name_list:
        print("  "+str(sl_no)+"\t"+str(col))
        sl_no+=1
    print()


def describe_gff_info(connection, args):
    cur = connection.cursor()
    query="SELECT * FROM summary"
    cur.execute(query)
    rows = cur.fetchone()
    col_name_list = [tuple[0] for tuple in cur.description]
    sl_no=1
    print("\nSerial\tColumn names\tDescription")
    for i, col in enumerate(col_name_list):
        desc_g = rows[i]
        if (desc_g ==""):
            desc_g = "--"
        print("  "+str(sl_no)+"\t"+str(col)+"\t\""+str(desc_g)+"\"")
        sl_no+=1
    print()


def stats_isomiR_per_miRNA(connection, miRNA_name, args):
    cur = connection.cursor()
    miR_array = add_mirnas(args)
    #cur.execute('SELECT * FROM data_sets WHERE seqID=?', (miRNA_name,))
    query="SELECT COUNT(*) FROM data_sets WHERE seqID=? AND type='isomiR' "
    query = add_filter(query, args)
    print()
    stat_counts=0
    if args.txtout:
        print("The results are being fetched and formated to be written to "+ args.txtout)
        #format_results()
        #with open(args.txtout, 'w') as w_stat:
        w_stat = open(args.txtout, 'w')
        w_stat.write("Serial number\tmiRNA\tisomiR Count\n")
    else:
        print("OUTPUT:")
    
    for miRs in miR_array:
        t=(miRs, )
        cur.execute(query, t)
        #cur.execute("SELECT COUNT(*) FROM data_sets WHERE seqID=? AND type='isomiR'", t)
        rows = cur.fetchall()
        for row in rows:
            stat_counts+=1
            row = row[0]
            if args.txtout:
                w_stat.write(str(stat_counts) +"\t"+miRs+"\t"+str(row)+"\n")
            else:
                print(str(stat_counts) +". " +"isomiRs for miRNA "+ miRs + ": "+ str(row))
    print()
    if args.txtout:
        w_stat.close()
    pass


def WHERE_CLAUSE(query, args):
    if "WHERE" in query: 
        query = query + " AND "
        return(query)
    else:
        query = query + " WHERE " 
        return(query)


# ALWAYS EXECUTE THIS LIMIT FUNCTION AT THE END 
def add_limit(query, args):
    if args.limit:
        query = query + " LIMIT "+ args.limit 
        return query
    else:
        return query


def add_filter(query, args):
    if args.filter:
        query = WHERE_CLAUSE(query, args)
        query = query + " filter='" + args.filter +"' "
        return query
    else:
        return query


def add_variants(query, args):
    my_var_dict = {'iso_5p': 'iso_5p != "None"', 'iso_3p': 'iso_3p != "None"', 'iso_add3p': 'iso_add3p != "None"', 'iso_add5p': 'iso_add5p != "None"', 
            'iso_snv_seed':'iso_snv_seed != 0', 'iso_snv_central_offset':'iso_snv_central_offset != 0', 'iso_snv_central':'iso_snv_central != 0', 
            'iso_snv_central_supp':'iso_snv_central_supp != 0', 'iso_snv':'iso_snv != 0'}
    user_req_var = args.variant.split(',')
    values_req_var =[]
    for eachVar in user_req_var:
        try:
            values_req_var.append(my_var_dict[eachVar])
        except KeyError:
            print("\nError: \"" + eachVar + "\" does not exist in the choices supported by (-var , --variant)\n")
            print("use: mirtop sql -qh for more options")
            exit()
    #print(values_req_var)
    insert_betwn = " AND "
    query_suffix = (insert_betwn.join( values_req_var ))
    query = WHERE_CLAUSE(query, args)
    query = query + query_suffix
    return query
    #if args.filter:
        #query = query + " AND " + query_suffix 
        #return query
    #else:
        #query = query + "WHERE " + query_suffix 
        #return query

def add_mirnas(args):
    if args.miRNA.endswith('.txt'):
        #print("I am called and I am safe here to read from a file")
        with open(args.miRNA, 'r') as miList:
            miR_array = miList.read().splitlines()
        if args.miRNA_prefix:
            element = str(args.miRNA_prefix) + "-"
            miR_array = [element + s  for s in miR_array]
            return(miR_array)
        else:
            return(miR_array)
    else:
        miR_array=args.miRNA.split(',')
        if args.miRNA_prefix:
            element = str(args.miRNA_prefix) + "-"
            miR_array = [element + s  for s in miR_array]
            return(miR_array)
        else:
            return(miR_array)


def perform_execution(conn, query, args):
    cur = conn.cursor()
    #print("QUERY: \n"+ query + "\n")
    cur.execute(query)
    rows = cur.fetchall()
    col_name_list = [tuple[0] for tuple in cur.description]
    if args.miRNA:
        return(col_name_list, rows)
    else:
        format_results(col_name_list, rows, args)

def format_results(header, output, args):
    header = '\t'.join(str(col) for col in header)
    if args.txtout:
        outList = open(args.txtout, 'w')
        print("\nWriting data to file: "+ args.txtout + "\n")
        outList.write(header+"\n")
        write_to_file(output, args, outList)
    else:
        print(header)
        if args.count:
            output = list(output[0])
            if args.columns:
                print("Unique counts for "+ str(args.columns) + " is: " + str(output[0]))
            else:
                print("Unique counts for all rows is: " + str(output[0]))
        else:
            for eachrow in output:
                row_tab = '\t'.join(str(items) for items in eachrow)
                print(row_tab)


def write_to_file(output, args, fileHandler):
        if args.miRNA:
            fileHandler.write(output + "\n")
        elif args.count:
            output = list(output[0])
            fileHandler.write("Unique counts for "+ args.columns + " is:\t" + str(output))
        else:
            for eachrow in output:
                row_tab = '\t'.join(str(items) for items in eachrow)
                fileHandler.write(row_tab+"\n")


def select_query(connection, args):
    if args.columns:
        if args.count: 
            if args.count == "T":
                query = "SELECT COUNT(" + args.columns + ") FROM data_sets "
            else:
                print("\nERROR: -n is incorrect!. \nPlease use -n T and optionally specify any one column in -col.\nFor more options see mirtop sql -h")
                exit()
        else:
            query = "SELECT " + args.columns + " FROM data_sets "
    elif args.count:
        query = "SELECT COUNT(*) FROM data_sets "
    else:
        query = "SELECT * FROM data_sets "
    query = add_filter(query, args)
    if args.variant:
        query = add_variants(query, args)
    if args.miRNA:
        miR_array = add_mirnas(args)
        query = WHERE_CLAUSE(query, args)
        query_series = query +  "seqID= "
        header_var= ""
        j=0
        if args.txtout:
            outList = open(args.txtout, 'w')
            print("\nWriting data to file: "+ args.txtout + "\n")

        for miRs in miR_array:
            query = query_series + "\"" + miRs + "\" "
            query = add_limit(query, args)
            (header, rows) = perform_execution(connection, query, args)
            j += 1
            for i, row in enumerate(rows):
                if (i == 0):
                    if header_var == "":
                        header_var = '\t'.join(str(col) for col in header)
                        if args.txtout:
                            outList.write(header_var+"\n")
                        else:
                            print(header_var)
                row_tab = '\t'.join(str(items) for items in row)
                if args.count:
                    newOut = str(j) + ". "+ miRs + ":\t" + row_tab
                    if args.txtout:
                        write_to_file(newOut, args, outList)
                    else:
                        print(newOut)
                else:
                    if args.txtout:
                        write_to_file(row_tab, args, outList)
                    else:
                        print(row_tab)
    else:
        query = add_limit(query, args)
        perform_execution(connection, query, args)


def sql_options(args):
    user_options = vars(args)
    if args.create:
        if args.gff:
            insert_sql(args)
        else:
            print("Usage: mirtop sql --create --gff <input.gff> --db <new_db_name> \(Default: mirtop.db\)")
    elif args.query:
        if args.expr:
            #  print("Usage: mirtop sql --query --db <input_database> -e <user_query>")
            query_sql(args)
        else:
            print("Usage: mirtop sql --query --db <input_database> -e <user_query>")
    else:
        print("Usage: mirtop sql -h")



"""mirGFF3 proxy converter"""
from __future__ import print_function

import os.path as op

from mirtop.mirna import fasta, mapper
from mirtop.bam.bam import read_bam
from mirtop.importer import seqbuster, srnabench, prost, isomirsea, manatee, optimir
from mirtop.mirna.annotate import annotate
from mirtop.gff import body, header, merge, read
from mirtop.mirna.mapper import read_gtf_to_mirna
import mirtop.libs.logger as mylog
logger = mylog.getLogger(__name__)


def reader(args):
    """
    Realign BAM hits to miRBAse to get better accuracy and annotation
    """
    if args.low_memory:
        read.reader(args)
        return None
    samples = []
    if args.database is None:
        database = mapper.guess_database(args)
    else:
        database = args.database
    args.database = database
    precursors = fasta.read_precursor(args.hairpin, args.sps)
    args.precursors = precursors
    matures = mapper.read_gtf_to_precursor(args.gtf, database)
    args.matures = matures
    # TODO check numbers of miRNA and precursors read
    # TODO print message if numbers mismatch
    out_dts = dict()
    if args.keep_name and len(args.files) > 1:
        logger.warning("--keep-name when running multiple samples\n"
                       "can generate wrong results if the\n"
                       "name read is different across sample\n"
                       "for the same sequence.")
    for fn in args.files:
        fn = op.normpath(fn)
        if args.format != "gff":
            sample = op.splitext(op.basename(fn))[0]
            samples.append(sample)
            fn_out = op.join(args.out, sample + ".%s" % args.out_format)
        if args.format == "BAM":
            reads = _read_bam(fn, args)
        elif args.format == "seqbuster":
            reads = seqbuster.read_file(fn, args)
        elif args.format == "srnabench":
            out_dts[fn] = srnabench.read_file(fn, args)
        elif args.format == "prost":
            reads = prost.read_file(fn, precursors, database, args.gtf)
        elif args.format == "isomirsea":
            out_dts[fn] = isomirsea.read_file(fn, args)
        elif args.format == "manatee":
            out_dts[fn] = manatee.read_file(fn, database, args)
        elif args.format == "optimir":
            out_dts[fn] = optimir.read_file(fn, args)
        elif args.format == "gff":
            samples.extend(header.read_samples(fn))
            out_dts[fn] = body.read(fn, args)
            continue
        if args.format not in ["isomirsea", "srnabench", "manatee", 'optimir']:
            ann = annotate(reads, matures, precursors)
            out_dts[fn] = body.create(ann, database, sample, args)
        h = header.create([sample], database, header.make_tools(args.format))
        _write(out_dts[fn], h, fn_out, args)
    # merge all reads for all samples into one dict
    if args.low_memory:
        return None
    merged = merge.merge(out_dts, samples)
    fn_merged_out = op.join(args.out, "%s.%s" % (args.prefix, args.out_format))
    _write(merged, header.create(samples, database, header.make_tools([args.format])), fn_merged_out, args)


def _write(lines, header, fn, args = None):
    out_handle = open(fn, 'w')
    print(header, file=out_handle)
    database = mapper.guess_database(args)
    dbmapper = read_gtf_to_mirna(args.gtf, database)
    for m in lines:
        for s in sorted(lines[m].keys()):
            for hit in lines[m][s]:
                # TODO: convert to genomic if args.out_genomic
                if args and args.out_genomic:
                    lifted = body.lift_to_genome(hit[4], dbmapper)
                    print(lifted, file=out_handle)
                else:
                    print(hit[4], file=out_handle)
    out_handle.close()


def _read_bam(bam_fn, precursors):
    if bam_fn.endswith("bam") or bam_fn.endswith("sam"):
        logger.info("Reading %s" % bam_fn)
        reads = read_bam(bam_fn, precursors)
    else:
        raise ValueError("Format not recognized."
                         " Only working with BAM/SAM files.")
    return reads



"""Update gff3 files to newest version"""
from __future__ import print_function
import sys
import os

from mirtop.gff.classgff import feature
from mirtop.gff.header import read_version, get_gff_version
from mirtop.mirna.keys import *
from mirtop.mirna.realign import make_id

import mirtop.libs.logger as mylog

logger = mylog.getLogger(__name__)


def convert(args):
    """
    Update previous GFF3 versions.

    Args:
        *args (namedtupled)*: arguments parsed from command line with
            *mirtop.libs.parse.add_subparser_update()*.

    Returns:
        *(out_file)*: most updated GFF3 file.
    """    
    for fn in args.files:
        out_fn = os.path.join(args.out, os.path.basename(fn))
        update_file(fn, out_fn)


def read_uid_10(idu):
    seq = ""
    for i in idu:
        if i == "1" or i == "2":
            return seq[:-int(i)]
        else:
            if i not in CODE2NT:
                logger.error("UID is not valid (%s)" % idu)
                return False
            seq += CODE2NT[i]
    return seq


def to10to11(gff_line):
    gff_line = gff_line.replace("_snp", "_snv")
    gff_line = gff_line.replace("_add", "_add3p")
    features = feature(gff_line)
    if "iso_5p" in features.attributes["Variant"]:
        variants = features.attributes["Variant"].split(",")
        iso_5p = [v.split(":") for v in variants if v.startswith("iso_5p")]
        iso_5p = -1 * int(iso_5p[0][1])
        if iso_5p > 0:
            iso_5p = "+%s" % iso_5p
        variants = ["iso_5p:%s" % iso_5p if v.startswith("iso_5p") else v for v in variants]
        features.attributes["Variant"] = ",".join(variants)
    features.attributes["UID"] = make_id(read_uid_10(features.attributes["UID"]))
    return features.paste_columns()


def update_file(gff_file, new_gff_file):
    """Update file from file version to current version"""
    versions = ["1.0"]
    functions = [to10to11]

    version = read_version(gff_file)
    init = versions.index(version)

    outh = sys.stdout if not new_gff_file else open(new_gff_file, 'w')
    with open(gff_file) as inh:
        for line in inh:
            if line.find("VERSION") > -1:
                print(get_gff_version(), file=outh)
                continue
            if line.startswith("##"):
                print(line.strip(), file=outh)
                continue
            for updates in range(init, len(functions)):
                print(functions[updates](line), file=outh)



import re
from Bio import pairwise2
from Bio.Seq import Seq
from collections import defaultdict

from mirtop.mirna.mintplates import convert
import mirtop.libs.logger as mylog

logger = mylog.getLogger(__name__)


class hits:
    """"Class with alignment information."""

    def __init__(self):
        self.sequence = ""
        self.idseq = ""
        self.precursors = defaultdict(isomir)
        self.score = []
        self.best_hits = []  # maybe sam object?
        self.counts = 0

    def set_sequence(self, seq):
        self.sequence = seq
        self.idseq = make_id(seq)

    def set_precursor(self, precursor, isomir):
        self.precursors[(precursor, isomir.start)] = isomir

    def remove_precursor(self, precursor):
        del self.precursors[precursor]


class isomir:
    """
    Class to represent isomiRs information.
    """

    def __init__(self):
        self.t5 = []
        self.t3 = []
        self.add = []
        self.subs = []
        self.external = ""
        self.align = None
        self.cigar = None
        self.filter = "Pass"
        self.map_score = 0
        self.end = None
        self.start = None
        self.mirna = None
        self.strand = "+"

    def set_pos(self, start, l, strand="+"):
        """Set end position"""
        if start < 0:
            # l = l + start
            start = 0
        self.strand = strand
        self.start = start
        self.end = start + l - 1
        if strand == "-":
            self.start = start + l - 1
            self.end = start

    def formatGFF(self):
        """Create Variant attribute."""
        value = []
        subs = self.subs
        if self.external != "notsure" and self.external != "":
            return self.external
        for sub in subs:
            if sub:
                if sub[0] > 1 and sub[0] < 8:
                    value.append("iso_snv_seed")
                elif sub[0] == 8:
                    value.append("iso_snv_central_offset")
                elif sub[0] > 8 and sub[0] < 13:
                    value.append("iso_snv_central")
                elif sub[0] > 12 and sub[0] < 18:
                    value.append("iso_snv_central_supp")
                else:
                    value.append("iso_snv")

        if self.add:
            value.append("iso_add3p:%s" % len(self.add))
        if self.t5:
            size = len(self.t5)
            direction = "-" if self.t5.isupper() else "+"
            value.append("iso_5p:%s%s" % (direction, size))
        if self.t3:
            size = len(self.t3)
            direction = "+" if self.t3.isupper() else "-"
            value.append("iso_3p:%s%s" % (direction, size))
        if not value:
            value = ["NA"]
        return ",".join(sorted(list(set(value))))

    def format(self, sep="\t"):
        """Create tabular line from variant fields."""
        subs = "".join(["".join(map(str, mism)) for mism in self.subs])
        if not subs:
            subs = "0"
        add = "0" if not self.add else self.add
        return "%s%s%s%s%s%s%s" % (subs, sep, add, sep,
                                   self.t5, sep, self.t3)

    def format_id(self, sep="\t"):
        """Create simple identifier from variant fields."""
        subs = ["".join(["".join([c[2], str(c[0]), c[1]]) for c in self.subs])]
        if not subs:
            subs = []
        add = [] if not self.add else ["e%s" % self.add]
        t5 = ["s%s" % self.t5] if self.t5 and self.t5 != "NA" else []
        t3 = ["%s" % self.t3] if self.t3 and self.t3 != "NA" else []
        full = t5 + subs + t3 + add
        return sep.join([f for f in full if f])

    def get_score(self, sc):
        """Get score from variant fields."""
        for a in self.add:
            if a in ['A', 'T']:
                sc -= 0.25
            else:
                sc -= 0.75
        for e in self.subs:
            sc -= 1
        return sc

    def is_iso(self):
        """Define whether element is isomiR or not."""
        if self.external == "NA":
            return False
        if self.t5 or self.t3 or self.add or self.subs or self.external != "":
            return True
        return False


def read_id(idu):
    """
    Read a unique identifier for the sequence and
    convert it to the nucleotides,
    replacing an unique code for 5 nts.

    It uses the code from *mirtop.mirna.keys()*.

    Inspired by MINTplate: https://cm.jefferson.edu/MINTbase
    https://github.com/TJU-CMC-Org/MINTmap/tree/master/MINTplates

    Args:
        *idu(str)*: unique identifier for the sequence.

    Returns:
        *seq(str)*: nucleotides sequences.
    """
    try:
        seq = convert(idu, False, 'iso')
    except KeyError:
        logger.error("UID is not valid " + idu)
        return False

    return seq


def make_id(seq):
    """
    Create a unique identifier for the sequence from the nucleotides,
    replacing 5 nts for a unique sequence.

    It uses the code from *mirtop.mirna.keys()*.

    Inspired by MINTplate: https://cm.jefferson.edu/MINTbase
    https://github.com/TJU-CMC-Org/MINTmap/tree/master/MINTplates

    Args:
        *seq(str)*: nucleotides sequences.

    Returns:
        *idName(str)*: unique identifier for the sequence.
    """
    try:
        idu = convert(seq, True, 'iso')
    except KeyError as error:
        logger.error("Sequence is not valid " + seq)
        raise

    # If you wanted to add "iso-" into the license plate as the prefix
    # idu = convert(seq, True, "iso")
    return idu


def is_sequence(seq):
    """
    This function check whether the sequence is valid or not.

    Args:
        *seq(str)*: string acting as a sequence.

    Returns:
        *boolean*: whether is or not a valid nucleotide sequence.
    """
    alphabet = re.compile('^[ACTG]*$', re.IGNORECASE)
    return alphabet.match(seq)


def align(x, y, local=False):
    """
    Pairwise alignments between two sequenes.
    https://medium.com/towards-data-science/pairwise-sequence-alignment-using-biopython-d1a9d0ba861f

    Args:
        *x(str)*: short sequence.

        *y(str)*: long sequence.

        *local(boolean)*: local or global alignment.

    Returns:
        *aligned_x(hit)*: alignment information, socre and positions.
    """
    if local:
        aligned_x = pairwise2.align.localxx(x, y)[0]
    else:
        aligned_x = pairwise2.align.globalms(x, y, 1, -1, -1, -0.5)[0]
    aligned_x = list(aligned_x)
    n_x = aligned_x[0]
    if "N" in n_x:
        N_indices = [i for i, ltr in enumerate(n_x) if ltr == 'N']
        n_x = list(n_x)
        for N_index in N_indices:
            n_x[N_index] = y[N_index]
        n_x = ''.join(n_x)

    aligned_x[0] = n_x
    return tuple(aligned_x)


def _add_cigar_char(counter, cache):
    if counter == 1:
        return cache
    else:
        return str(counter) + cache


def make_cigar(seq, mature):
    """
    Function that will create CIGAR string from aligment
    between read and reference sequence.

    Args:
        *seq(str)*: read sequence.

        *mature(str)*: short sequence.

    Return:
        *short(str)*: CIGAR string.
    """
    cigar = ""
    for pos in range(0, len(seq)):
        if seq[pos] == mature[pos]:
            cigar += "M"
        elif seq[pos] != mature[pos] and seq[pos] != "-" and mature[pos] != "-":
            cigar += seq[pos]
        elif seq[pos] == "-":
            cigar += "D"
        elif mature[pos] == "-":
            cigar += "I"

    cache = ""
    counter = 1
    short = ""
    for c in cigar:
        if c != cache and cache != "" and cache == "M":
            short += _add_cigar_char(counter, cache)
            counter = 1
            cache = c
        if c != "M":
            short += c
        if c == cache and c == "M":
            counter += 1
        cache = c

    if cache == "M":
        short += _add_cigar_char(counter, cache)
    return short


def cigar_correction(cigarLine, query, target):
    """
    Read from CIGAR in BAM file to define mismatches.

    Args:
        *cirgarLine(str)*: CIGAR string from BAM file.

        *query(str)*: read sequence.

        *target(str)*: target sequence.

    Returns:
        *(list)*: [query_nts, target_nts]
    """
    query_pos = 0
    target_pos = 0
    query_fixed = []
    target_fixed = []
    for (cigarType, cigarLength) in cigarLine:
        if cigarType == 0:  # match
            query_fixed.append(query[query_pos:query_pos+cigarLength])
            target_fixed.append(target[target_pos:target_pos+cigarLength])
            query_pos = query_pos + cigarLength
            target_pos = target_pos + cigarLength
        elif cigarType == 1:  # insertions
            query_fixed.append(query[query_pos:query_pos+cigarLength])
            target_fixed.append("".join(["-"] * cigarLength))
            query_pos = query_pos + cigarLength
        elif cigarType == 2:  # deletion
            target_fixed.append(target[target_pos:target_pos+cigarLength])
            query_fixed.append("".join(["-"] * cigarLength))
            target_pos = target_pos + cigarLength
    return ["".join(query_fixed), "".join(target_fixed)]


def expand_cigar(cigar):
    """
    From short CIGAR version to long CIGAR version
    where each character is each nts in the sequence.

    Args:
        *cigar(str)*: CIGAR string.

        >>> 10MA3M

    Returns:
        *cigar_long(str)*: CIGAR long.

        >>> MMMMMMMMMMAMMM
    """
    cigar_long = ""
    n = 0
    for nt in cigar:
        if nt in ["D", "M", "I", "A", "T", "C", "G"]:
            if n > 0:
                cigar_long += nt * int(n)
            else:
                cigar_long += nt
            n = 0
        else:
            if n > 0:
                n = int("%s%s" % (n, nt))
            else:
                n = int(nt)
    return cigar_long


def cigar2snp(cigar, reference):
    """
    From a CIGAR string and reference sequence
    detect mistmatches positions and reference and
    target nucleotides.

    Args:
        *cigar(str)*: CIGAR string.

        *reference(str)*: reference sequence.

    Returns:
        *snp(list)*: position of mismatches (indels included) as:

        >>> [pos, seq_nt, ref_nt]
    """
    snp = []
    pos_seq = 0
    pos_ref = 0
    for nt in expand_cigar(cigar):
        if nt != "M":
            if nt == "I":
                snp.append([pos_seq, nt, "-"])
                pos_seq += 1
            elif nt == "D":
                snp.append([pos_seq, "-", reference[pos_ref]])
                pos_ref += 1
            else:
                snp.append([pos_seq, nt, reference[pos_ref]])
                pos_ref += 1
                pos_seq += 1
        else:
            pos_ref += 1
            pos_seq += 1
    return snp


def reverse_complement(seq):
    """Get reverse complement of a sequences

    Args:
        *seq(str)*: sequence.

        >>> GCAT

    Returns:
        *(str)*: reverse complemente sequence:

        >>> ATGC
    """
    return str(Seq(seq).reverse_complement())


def get_mature_sequence(precursor, mature, exact=False, nt = 5):
    """
    From precursor and mature positions
       get mature sequence with +/- 4 flanking nts.

    Args:
       *precursor(str)*: long sequence.

       *mature(list)*: [start, end].

       *exact(boolean)*: not add 4+/- flanking nts.

       *nt(int)*: number of nts to get.

    Returns:
        *(str)*: mature sequence.
    """
    p = "%s%s" % ("".join(["N"]*nt), precursor)
    s = mature[0] + nt
    e = mature[1] + nt
    if exact:
        return p[s:e + 1]
    return p[s - (nt - 1) :e + nt]


def align_from_variants(sequence, mature, variants):
    """
    Giving the sequence read,
       the mature from get_mature_sequence,
       and the variant GFF annotation:
       get a list of substitutions

    Args:
        *sequence(str)*: read sequence.

        *mature(str)*: mature sequence from
            *mirtop.mirna.realing.get_mature_sequence()*.

        *variants(str)*: string from Variant attribute in GFF file.

    Returns:
        *snp(list)*: [[pos, target, reference]]
    """
    init_log = "iso:%s -> %s\nref:%s" % (sequence, variants, mature)
    snps = []
    k = [v.split(":")[0] for v in variants.split(",") if v.find(":") > -1]
    v = [int(v.split(":")[1]) for v in variants.split(",") if v.find(":") > -1]
    var_dict = dict(zip(k, v))
    logger.debug("realign::align_from_variants::sequence %s" % sequence)
    logger.debug("realign::align_from_variants::mature %s" % mature)
    logger.debug("realign::align_from_variants::variants %s" % variants)
    snp = ["iso_snv" for v in variants.split(",") if v.find("snv") > -1]
    fix_5p = 7
    if "iso_5p" in k:
        fix_5p = 7 + var_dict["iso_5p"]
    mature = mature[fix_5p:]
    if "iso_add3p" in k:
        sequence = sequence[:-1 * var_dict["iso_add3p"]]
    if "iso_3p" in k:
        shift = 7 + (-1 * var_dict["iso_3p"])
        if shift != 0:
            mature = mature[:-(shift)]
    else:
        mature = mature[:-7]
    logger.debug("realign::align_from_variants::snp %s" % snp)
    logger.debug("realign::align_from_variants::sequence %s" % sequence)
    logger.debug("realign::align_from_variants::mature %s" % mature)
    if len(sequence) != len(mature):  # in case of indels, align again
        a = align(sequence, mature)
        sequence = a[0]
        mature = a[1]

    if len(sequence) > len(mature):
        logger.warning("Invalid isomiR definition:\n%s\niso:%s\nref:%s" % (init_log, sequence, mature))
        return "Invalid"

    for p in range(0, len(sequence)):
        if sequence[p] != mature[p]:
            if mature[p] == "N":
                continue
            value = ""
            pos = p + 1
            value = "iso_snv"
            logger.debug("realign::align_from_variants::value %s at %s" % (value, pos))
            if value in snp:
                snps.append([pos, sequence[p], mature[p]])
    logger.debug("realign::align_from_variants::snps %s" % snps)
    return snps


def variant_to_5p(hairpin, pos, variant):
    """
    From a sequence and a start position get the nts
       +/- indicated by iso_5p. Pos option is 0-base-index

    Args:
       *hairpin(str)*: long sequence:
            >>> AAATTTT

       *position(int)*: >>> 3

       *variant(int)*: number of nts involved in the variant:
            >>> -1
    Returns:
       *(str)*: nucleotide involved in the variant:
            >>> T
    """
    pos = pos[0]
    iso_t5 = [v for v in variant.split(",") if v.startswith("iso_5p")]
    if iso_t5:
        t5 = int(iso_t5[0].split(":")[-1][-1])
        direction_t5 = int(iso_t5[0].split(":")[-1]) * -1
        if direction_t5 > 0:
            return hairpin[pos - t5:pos]
        elif direction_t5 < 0:
            return hairpin[pos:pos + t5].lower()
    return "0"


def variant_to_3p(hairpin, pos, variant):
    """
    From a sequence and a start position get the nts
       +/- indicated by iso_3p. Pos option is 0-base-index

    Args:
       *hairpin(str)*: long sequence:
            >>> AAATTTT

       *position(int)*: >>> 3

       *variant(int)*: number of nts involved in the variant:
            >>> -1
    Returns:
       *(str)*: nucleotide involved in the variant:
            >>> A
    """
    pos = pos[1]
    iso_t3 = [v for v in variant.split(",") if v.startswith("iso_3p")]
    if iso_t3:
        t3 = int(iso_t3[0].split(":")[-1][-1])
        direction_t3 = int(iso_t3[0].split(":")[-1])
        if direction_t3 > 0:
            return hairpin[pos + 1:pos + t3 + 1]
        elif direction_t3 < 0:
            return hairpin[pos - t3 + 1:pos + 1].lower()
    return "0"


def variant_to_add(read, variant):
    """
    From a sequence and a start position get the nts
       +/- indicated by iso_3p. Pos option is 0-base-index

    Args:
       *hairpin(str)*: long sequence:
            >>> AAATTTT

       *position(int)*: >>> 3

       *variant(int)*: number of nts involved in the variant:
            >>> 2
    Returns:
       *(str)*: nucleotide involved in the variant:
            >>> TT
    """
    add = [v for v in variant.split(",") if v.startswith("iso_add")]
    if add:
        add = int(add[0].split(":")[-1][-1]) * -1
        return read[add:]
    return "0"



""" Read isomiR GFF files"""

import os
from collections import defaultdict, Counter

import mirtop.libs.logger as mylog
from mirtop.mirna import mapper
from mirtop.mirna.realign import expand_cigar, make_id
from mirtop.gff.body import paste_columns, read_attributes
from mirtop.gff.body import variant_with_nt
from mirtop.gff.classgff import feature

logger = mylog.getLogger(__name__)


def header(fn):
    """
    Custom header for isomiR-SEA importer.

    Args:
        *fn (str)*: file name with isomiR-SEA GFF output

    Returns:
        *(str)*: isomiR-SEA header string.
    """
    h = ""
    return h


def read_file(fn, args):
    """
    Read isomiR-SEA file and convert to mirtop GFF format.

    Args:
        *fn(str)*: file name with isomiR-SEA output information.

        *database(str)*: database name.

        *args(namedtuple)*: arguments from command line.
            See *mirtop.libs.parse.add_subparser_gff()*.

    Returns:
        *reads (nested dicts)*:gff_list has the format as
            defined in *mirtop.gff.body.read()*.

    """
    database = args.database
    gtf = args.gtf
    sep = " " if args.out_format == "gtf" else "="
    map_mir = mapper.read_gtf_to_mirna(gtf)
    reads = defaultdict(dict)
    reads_in = 0
    sample = os.path.splitext(os.path.basename(fn))[0]
    hits = _get_hits(fn)
    logger.debug("ISOMIRSEA::SAMPLE::%s" % sample)
    with open(fn) as handle:
        for line in handle:
            cols = line.strip().split("\t")
            attr = read_attributes(line, "=")
            query_name = attr['TS']
            query_sequence = attr['TS'].replace("U", "T")
            start = int(cols[3])
            end = int(cols[4])
            isomirseq_iso = attr['ISO']
            if query_name not in reads and query_sequence == None:
                continue
            if query_sequence and query_sequence.find("N") > -1:
                continue
            counts = attr["TC"]
            chrom = cols[0]
            # logger.debug("SEQBUSTER:: cigar {cigar}".format(**locals()))
            cigar = attr['CI'].replace("U", "T")
            idu = make_id(query_sequence)
            isoformat = cigar2variants(cigar, query_sequence, attr['ISO'])
            logger.debug("\nISOMIRSEA::NEW::query: {query_sequence}\n"
                         "  precursor {chrom}\n"
                         "  name: {query_name}\n"
                         "  idu: {idu}\n"
                         "  start: {start}\n"
                         "  cigar: {cigar}\n"
                         "  iso: {isoformat}\n"
                         "  variant: {isoformat}".format(**locals()))
            source = "isomiR" if isoformat != "NA" else "ref_miRNA"
            strand = "+"
            database = cols[1]
            mirName = attr['MIN'].split()[0]
            preName = attr['PIN'].split()[0]
            score = "."
            Filter = attr['FILTER']
            isotag = attr['ISO']
            tchrom, tstart = _genomic2transcript(map_mir[mirName],
                                                 chrom, start)
            start = start if not tstart else tstart
            chrom = chrom if not tstart else tchrom
            end = start + len(query_sequence)
            hit = hits[idu]
            fields = {'seq_name': query_sequence, 'idseq': idu,
                      'name': mirName, 'parent': preName,
                      'variant': isoformat, 'cigar': cigar,
                      'counts': counts, 'filter': Filter,
                      'hits': hit, 'chrom': chrom,
                      'start': start, 'end': end,
                      'database': database, 'source': source,
                      'score': score, 'strand': strand}
            # TODO: convert to genomic if args.out_genomic
            line = feature(fields).line
            if args.add_extra:
                extra = variant_with_nt(line, args.precursors, args.matures)
                line = "%s Changes %s;" % (line, extra)

            line = paste_columns(feature(line), sep=sep)
            if start not in reads[chrom]:
                reads[chrom][start] = []
            if Filter == "Pass":
                reads_in += 1
                reads[chrom][start].append([idu, chrom, counts, sample, line])

    logger.info("Hits: %s" % reads_in)
    return reads


def _get_hits(fn):
    hits = Counter()
    with open(fn) as handle:
        for line in handle:
            attr = read_attributes(line, "=")
            query_sequence = attr['TS'].replace("U", "T")
            if query_sequence and query_sequence.find("N") > -1:
                continue
            idu = make_id(query_sequence)
            hits[idu] += 1
    return hits


def cigar2variants(cigar, sequence, tag):
    """From cigar to Variants in GFF format"""
    pos = 0
    iso5p = 0
    logger.debug("\nISOMIRSEA:: expanded: %s" % expand_cigar(cigar))
    for l in expand_cigar(cigar):
        if l == "I":
            iso5p -= 1
        elif l == "D":
            iso5p += 1
        else:
            break
    iso3p = 0
    for l in reversed(expand_cigar(cigar)):
        if l == "I":
            iso3p += 1
        elif l == "D":
            iso3p -= 1
        else:
            break
    isosnp = []
    for l in expand_cigar(cigar):
        if l in ['A', 'T', 'C', 'G' ]:
            isosnp.append([pos, sequence[pos], l])
        if l in ['D']:
            continue
        pos += 1
    iso5p = "iso_5p:%s" % _fix(iso5p) if iso5p else ""
    if tag[-1] == "T" or iso3p < 0:
        iso3p = "iso_3p:%s" % _fix(iso3p) if iso3p else ""
    else:
        iso3p = "iso_add3p:%s" % iso3p if iso3p else ""

    variant = ""
    for iso in [iso5p, iso3p, _define_snp(isosnp)]:
        if iso:
            variant += "%s," % iso

    variant = "NA;" if not variant else variant
    return variant[:-1]


def _define_snp(subs):
    value = ""
    logger.debug("\nISOMIRSEA:: subs %s" % subs)
    for sub in subs:
        if sub:
            if sub[0] > 1 and sub[0] < 8:
                value += "iso_snv_seed,"
            elif sub[0] == 8:
                value += "iso_snv_central_offset,"
            elif sub[0] > 8 and sub[0] < 13:
                value += "iso_snv_central,"
            elif sub[0] > 12 and sub[0] < 18:
                value += "iso_snv_central_supp,"
            else:
                value += "iso_snv,"
    return value[:-1]


def _fix(n):
    if n > 0:
        return "+%s" % n
    return n


def _genomic2transcript(code, chrom, pos):
    for ref in code:
        if _is_chrom(chrom, code[ref][0]):
            if _is_inside(pos, code[ref][1:3]):
                return [ref, _transcript(pos, code[ref][1:4])]
    return [None, None]


def _is_chrom(chrom, annotated):
    logger.debug("TRANSCRIPT::CHROM::read position %s and db position %s" % (chrom, annotated))
    if chrom == annotated:
        return True
    if chrom == annotated.replace("chr", ""):
        return True
    return False


def _is_inside(pos, annotated):
    logger.debug("TRANSCRIPT::INSIDE::read position %s and db position %s" % (pos, annotated))
    if pos > annotated[0] and pos < annotated[1]:
        return True
    return False


def _transcript(pos, annotated):
    logger.debug("TRANSCRIPT::TRANSCRIPT::read position %s and db position %s" % (pos, annotated))
    if annotated[2] == "+":
        return pos - annotated[0]
    elif annotated[2] == "-":
        return annotated[1] - pos
    raise ValueError("Strand information is incorrect %s" % annotated[3])



""" Read Manatee files"""
from __future__ import print_function

from collections import defaultdict
import os

import mirtop.libs.logger as mylog
from mirtop.bam.bam import intersect
from mirtop.bam import filter
from mirtop.mirna.mapper import get_primary_transcript, guess_database
from mirtop.mirna.realign import isomir, reverse_complement, make_id, hits
from mirtop.gff.body import paste_columns, variant_with_nt
# from mirtop.mirna import mapper
from mirtop.gff.classgff import feature
from mirtop.gff import body
from mirtop.mirna.annotate import annotate

logger = mylog.getLogger(__name__)


def read_file(fn, database, args):
    """
    Read Manatee file and convert to mirtop GFF format.

    Args:
        *fn(str)*: file name with Manatee output information.

        *database(str)*: database name.

        *args(namedtuple)*: arguments from command line.
            See *mirtop.libs.parse.add_subparser_gff()*.

    Returns:
        *reads (nested dicts)*:gff_list has the format as
            defined in *mirtop.gff.body.read()*.

    """
    reads = defaultdict(dict)
    sample = os.path.splitext(os.path.basename(fn))[0]
    precursors = args.precursors
    bed_fn = os.path.join(args.out, os.path.basename(fn) + ".bed")
    sep = " " if args.out_format == "gtf" else "="
    seen = set()
    with open(fn, 'r') as handle:
        if not os.path.exists(bed_fn):
            _bed(handle, bed_fn)
    intersect_fn = intersect(bed_fn, args.gtf)
    for line in intersect_fn:
        data = _analyze_line(line, precursors, database, sample, sep, args)
        if data:
            start = data["start"]
            chrom = data["chrom"]
            key = "%s:%s" % (data['mirna'], data["name"])
            if start not in reads[chrom]:
                reads[chrom][start] = []
            if key not in seen:
                seen.add(key)
                reads[chrom][start].append(data["line"])
    return reads


def _analyze_line(line, precursors, database, sample, sep, args):
    start_idx = 10
    end_idx = 11
    attr_idx = 15
    query_name = line[3]
    sequence = line[4]
    if str(line).find(get_primary_transcript(guess_database(args))) < 0: # only working with mirbase
        return None

    logger.debug(("READ::line name:{0}").format(line))
    if sequence and sequence.find("N") > -1:
        return None

    chrom = line[attr_idx].strip().split("Name=")[-1]
    start = line[1]
    end = line[2]
    strand = line[5]
    counts = float(line[6])
    Filter = "Pass"
    reads = dict()
    if not start:
        return None
    if strand == "+":
        start = int(start) - int(line[start_idx]) + 1
    else:
        start = int(line[end_idx]) - int(end)
    iso = isomir()
    iso.align = line
    iso.set_pos(start, len(sequence))
    logger.debug("READ::From BAM start %s end %s at chrom %s" % (iso.start, iso.end, chrom))
    if len(precursors[chrom]) < start + len(sequence):
        logger.debug("READ::%s start + %s sequence size are bigger than"
                     " size precursor %s" % (
                                             chrom,
                                             len(sequence),
                                             len(precursors[chrom])))
    iso.subs, iso.add, iso.cigar = filter.tune(
        sequence, precursors[chrom],
        start, None)
    logger.debug("READ::After tune start %s end %s" % (iso.start, iso.end))
    logger.debug("READ::iso add %s iso subs %s" % (iso.add, iso.subs))

    idu = make_id(sequence)
    reads[query_name] = hits()
    reads[query_name].set_sequence(sequence)
    reads[query_name].counts = counts
    reads[query_name].sequence = sequence
    reads[query_name].set_precursor(chrom, iso)
    reads = annotate(reads, args.matures, args.precursors, quiet=True)
    gff_line = body.create(reads, args.database, sample, args, quiet=True)
    if start not in gff_line[chrom]:
        return None
    line = gff_line[chrom][start][0][4]
    logger.debug("READ::line:%s" % line)
    if args.add_extra:
        extra = variant_with_nt(line, args.precursors,
                                args.matures)
        line = "%s Changes %s;" % (line, extra)

    line = paste_columns(feature(line), sep=sep)
    return {'chrom': chrom,
            'start': start,
            'name': query_name,
            'mirna': reads[query_name].precursors[chrom].mirna,
            'line': [idu, chrom, counts, sample, line]}


def _bed(handle, bed_fn):
    with open(bed_fn, 'w') as outh:
        for line in handle:
            if line.startswith("@"):
                continue
            cols = line.strip().split()
            if cols[2]=="*":
                logger.debug("READ::Sequence not mapped: %s" % cols[0])
                continue
            query_name = cols[0]
            query_sequence = cols[9]
            counts = cols[14]
            start = int(cols[3])
            strand = "-" if cols[1] == "16" else "+"
            chrom = cols[2]
            # if there no hits
            # if the sequence always matching the read, assuming YES now
            # if not current or query_name!=current:
            query_sequence = query_sequence if not strand=="-" else reverse_complement(query_sequence)
            # logger.debug(("READ::Read name:{0} and Read sequence:{1}").format(line.query_name, sequence))
            if query_sequence and query_sequence.find("N") > -1:
                continue
            end = start + len(query_sequence) - 1
            bed_line = "\t".join(list(map(str, [chrom, start, end, query_name,
                                                query_sequence, strand, counts])))
            outh.write(bed_line + '\n')



from __future__ import print_function

from optparse import OptionParser
import os
import random
import numpy

from mirtop.mirna import fasta
from mirtop.mirna import mapper
from mirtop.mirna import realign
from mirtop.gff import body, header
import mirtop.libs.logger as mylog
logger = mylog.getLogger(__name__)


def write_collapse_fastq(reads, out_fn):
    idx = 0
    with open(out_fn, 'a') as outh:
        for r in reads:
            idx += 1
            print(">name%s_x%s" % (idx, r[1]), file=outh)
            print(r[0], file=outh)


def write_fastq(reads, out_fn):
    idx = 0
    with open(out_fn, 'a') as outh:
        for r in reads:
            idx += 1
            print("@name_read:%s" % idx, file=outh)
            print(r, file=outh)
            print("+", file=outh)
            print("I" * len(r), file=outh)


def create_read(read, count, adapter="TGGAATTCTCGGGTGCCAAGGAACTC", size=36):
    reads = list()
    for i in range(0, count):
        rest = size - len(read)
        part = adapter[:rest]
        reads.append(read + part)
    return reads


def variation(info, seq):
    randS = random.randint(info[0] - 2, info[0] + 2) + 1
    randE = random.randint(info[1] - 1, info[1] + 2) + 1
    if randS < 1:
        randS = 1
    if randE > len(seq):
        randE = info[1] - 1
    randSeq = seq[randS:randE]
    t5Lab = ""
    t5Lab = seq[randS:info[0]] if randS < info[0] else t5Lab
    t5Lab = seq[info[0]:randS].lower() if randS > info[0] else t5Lab
    t3Lab = ""
    t3Lab = seq[randE:info[1] + 1].lower() if randE < info[1] + 1 else t3Lab
    t3Lab = seq[info[1] + 1:randE] if randE > info[1] + 1 else t3Lab
    # mutation
    isMut = random.randint(0, 10)
    mutLab = []
    if isMut == 3:
        ntMut = random.randint(0, 3)
        posMut = random.randint(0, len(randSeq) - 1)
        if not randSeq[posMut] == nt[ntMut]:
            temp = list(randSeq)
            mutLab = [[posMut, nt[ntMut], randSeq[posMut]]]
            temp[posMut] = nt[ntMut]
            randSeq = "".join(temp)

    # addition
    isAdd = random.randint(0, 3)
    addTag = ""
    if isAdd == 2:
        posAdd = random.randint(1, 3)
        for numadd in range(posAdd):
            ntAdd = random.randint(0, 1)
            print([randSeq, seq[randS + len(randSeq)]])
            if nt[ntAdd] == seq[randS + len(randSeq)]:
                ntAdd = 1 if ntAdd == 0 else 0
            randSeq += nt[ntAdd]
            addTag += nt[ntAdd]
            print([randSeq, randE, info[1]])
    return [randSeq, randS, t5Lab, t3Lab, mutLab, addTag]


def create_iso(name, mir, seq, numsim, exp):
    reads = dict()
    full_read = list()
    clean_read = list()
    seen = set()
    for mirna in mir[name]:
        info = mir[name][mirna]
        for rand in range(int(numsim)):
            e = 1
            if exp:
                trial = random.randint(1, 100)
                p = random.randint(1, 50) / 50.0
                e = numpy.random.negative_binomial(trial, p, 1)[0]
            iso = realign.isomir()
            randSeq, iso.start, iso.t5, iso.t3, iso.subs, iso.add = variation(info, seq)
            if randSeq in seen:
                continue
            seen.add(randSeq)
            iso.end = iso.start + len(randSeq)
            aln = realign.align(randSeq, seq[iso.start:iso.end])
            iso.cigar = realign.make_cigar(aln[0], aln[1])
            iso.mirna = mirna
            query_name = "%s.%s.%s" % (mirna, iso.format_id("."), randSeq)
            reads[query_name] = realign.hits()
            reads[query_name].set_sequence(randSeq)
            reads[query_name].counts = e
            reads[query_name].set_precursor(name, iso)
            full_read.extend(create_read(randSeq, e))
            clean_read.append([randSeq, e])
            # print([randSeq, mutLab, addTag, t5Lab, t3Lab, mirSeq])
            # data[randSeq] = [exp, iso] # create real object used in code to generate GFF
    write_fastq(full_read, full_fq)
    write_collapse_fastq(clean_read, clean_fq)
    gff = body.create(reads, "miRBase21", "sim1")
    return gff


def _write(lines, header, fn):
    out_handle = open(fn, 'w')
    print(header, file=out_handle)
    for m in lines:
        for s in sorted(lines[m].keys()):
            for hit in lines[m][s]:
                print(hit[4], file=out_handle)
    out_handle.close()


usagetxt = "usage: %prog  --fa precurso.fa --gtf miRNA.gtf -n 10"


parser = OptionParser(usage=usagetxt, version="%prog 1.0")
parser.add_option("--fa",
                  help="", metavar="FILE")
parser.add_option("--gtf",
                  help="", metavar="FILE")
parser.add_option("-n", "--num", dest="numsim",
                  help="")
parser.add_option("-e", "--exp", dest="exp", action="store_true",
                  help="give expression", default=False)
parser.add_option("-p", "--prefix", help="output name")
parser.add_option("--seed", help="set up seed for reproducibility.", default=None)


(options, args) = parser.parse_args()

if options.seed:
    random.seed(options.seed)

full_fq = "%s_full.fq" % options.prefix
clean_fq = "%s_clean.fq" % options.prefix
out_gff = "%s.gff" % options.prefix
if os.path.exists(full_fq):
    os.remove(full_fq)
if os.path.exists(clean_fq):
    os.remove(clean_fq)

pre = fasta.read_precursor(options.fa, "")
mir = mapper.read_gtf_to_precursor(options.gtf)

nt = ['A', 'T', 'G', 'C']
gffs = dict()
h = header.create(["sampleX"], "miRBase1", "")
for precursor in pre:
    seq = pre[precursor]
    gffs.update(create_iso(precursor, mir, seq, options.numsim, options.exp))


_write(gffs, h, out_gff)
